/*
86d, an 8086 disassembler

Written by Taylor Gurney, 2022-2024.
Public domain.

This file contains the public API for lib86d.

The library provides three facilities:

a86d_decode: parse machine code bytes into an a86d_ins structure,
which represents a single 8086 instruction.

a86d_disasm: convert an a86d_ins structure to an assembly
language string.

a86d_rel_to_abs: convert relative calls/jumps to absolute ones for
display purposes.
*/
#ifndef A86D_H
#define A86D_H

#include "bits.h"

enum {
    /* Minimum required length of the output buffer for the assembly
       language string (the outbuf parameter to a86d_disasm). */
    A86D_DISASM_OUTBUF_LEN = 64
};

/* The struct a86d_ins and all of the following enums are outputs from
   the decoder. */

/* What type of instruction it is. */
enum a86d_op {
    a86d_op_none = 0,
    a86d_op_aaa = 1,
    a86d_op_aad = 2,
    a86d_op_aam = 3,
    a86d_op_aas = 4,
    a86d_op_adc = 5,
    a86d_op_add = 6,
    a86d_op_and = 7,
    a86d_op_call_abs32 = 8, /* far call, immediate address */
    a86d_op_call_mem16 = 9, /* near call, address from register or memory */
    a86d_op_call_mem32 = 10, /* far call, address from memory */
    a86d_op_call_rel16 = 11, /* near call, 16-bit unsigned displacement */
    a86d_op_cbw = 12,
    a86d_op_clc = 13,
    a86d_op_cld = 14,
    a86d_op_cli = 15,
    a86d_op_cmc = 16,
    a86d_op_cmp = 17,
    a86d_op_cmpsb = 18,
    a86d_op_cmpsw = 19,
    a86d_op_cwd = 20,
    a86d_op_daa = 21,
    a86d_op_das = 22,
    a86d_op_dec = 23,
    a86d_op_div = 24,
    a86d_op_esc = 25,
    a86d_op_hlt = 26,
    a86d_op_idiv = 27,
    a86d_op_imul = 28,
    a86d_op_in = 29,
    a86d_op_inc = 30,
    a86d_op_int = 31,
    a86d_op_int3 = 32,
    a86d_op_into = 33,
    a86d_op_iret = 34,
    a86d_op_ja = 35,
    a86d_op_jbe = 36,
    a86d_op_jc = 37,
    a86d_op_jcxz = 38,
    a86d_op_jz = 39,
    a86d_op_jg = 40,
    a86d_op_jge = 41,
    a86d_op_jl = 42,
    a86d_op_jle = 43,
    a86d_op_jmp_abs32 = 44,
    a86d_op_jmp_mem16 = 45,
    a86d_op_jmp_mem32 = 46,
    a86d_op_jmp_rel16 = 47,
    a86d_op_jmp_rel8 = 48,
    a86d_op_jnc = 49,
    a86d_op_jnz = 50,
    a86d_op_jno = 51,
    a86d_op_jns = 52,
    a86d_op_jo = 53,
    a86d_op_jp = 54,
    a86d_op_jpo = 55,
    a86d_op_js = 56,
    a86d_op_lahf = 57,
    a86d_op_lds = 58,
    a86d_op_lea = 59,
    a86d_op_les = 60,
    a86d_op_lodsb = 61,
    a86d_op_lodsw = 62,
    a86d_op_loop = 63,
    a86d_op_loopnz = 64,
    a86d_op_loopz = 65,
    a86d_op_mov = 66,
    a86d_op_movsb = 67,
    a86d_op_movsw = 68,
    a86d_op_mul = 69,
    a86d_op_neg = 70,
    a86d_op_nop = 71,
    a86d_op_not = 72,
    a86d_op_or = 73,
    a86d_op_out = 74,
    a86d_op_pop = 75,
    a86d_op_popf = 76,
    a86d_op_push = 77,
    a86d_op_pushf = 78,
    a86d_op_rcl = 79,
    a86d_op_rcr = 80,
    a86d_op_ret = 81,
    a86d_op_retf = 82,
    a86d_op_rol = 83,
    a86d_op_ror = 84,
    a86d_op_sahf = 85,
    a86d_op_salc = 86,
    a86d_op_sar = 87,
    a86d_op_sbb = 88,
    a86d_op_scasb = 89,
    a86d_op_scasw = 90,
    a86d_op_shl = 91,
    a86d_op_shr = 92,
    a86d_op_stc = 93,
    a86d_op_std = 94,
    a86d_op_sti = 95,
    a86d_op_stosb = 96,
    a86d_op_stosw = 97,
    a86d_op_sub = 98,
    a86d_op_test = 99,
    a86d_op_wait = 100,
    a86d_op_xchg = 101,
    a86d_op_xlat = 102,
    a86d_op_xor = 103,
    a86d_op_db = 104    /* Raw bytes, not an instruction. */
};

/* What rep prefix is present, if any.
   a86d_no_rep: No rep prefix.
   a86d_rep: rep/repe/repz
   a86d_repne: rep/repne/repnz
*/
enum a86d_rep_mode {
    a86d_no_rep = 0,
    a86d_rep = 1,
    a86d_repne = 2
};

/* Segment override prefix. In the a86d_ins struct, this is always populated.
   If the instruction has no segment override then it defaults to ds.
*/
enum {
    a86d_seg_es = 0,
    a86d_seg_cs = 1,
    a86d_seg_ss = 2,
    a86d_seg_ds = 3,
    a86d_seg_none = 4
};

/* Specifies the type of each operand. */
enum a86d_oper_type {
    a86d_oper_none = 0,      /* no operand */
    a86d_oper_reg = 1,       /* example: al */
    a86d_oper_segreg = 2,    /* example: ds */
    a86d_oper_imm = 3,       /* example: 0x1234 */
    a86d_oper_relimm = 4,    /* example: 0x1234 */
    a86d_oper_mem = 5,       /* example: [bx + si + 0x1234] */
    a86d_oper_directaddr = 6 /* example: [0x1234] */
};

/* This provides all information necessary for an emulated 8086 to
   execute the instruction, or to convert it to assembly language.
*/
struct a86d_ins {

    /* 1 if the instruction has a lock prefix. 0 if not. */
    int lock;

    /* rep prefix */
    enum a86d_rep_mode rep_mode;

    /* what type of instruction it is. */
    enum a86d_op op;

    /* segment override */
    byte seg;

    /* Number of bytes that were decoded for this instruction. 8086
       instructions can be arbitrary length due to prefixes, so if you
       want the full byte string you'll have to keep track of it
       yourself in the next_byte function you provide to the decoder.
     */
    dword length;

    /* The type of each operand. If there is only one operand, then the
       second element of this array will be a86d_oper_none. If there are
       no operands, then both elements of this array will be a86d_oper_none.
     */
    enum a86d_oper_type optype[2];

    /* Both of these array elements are always present, but if the
       corresponding optype is a86d_oper_none, then the value of that
       operand is undefined and meaningless.
     */
    dword operand[2];

    /* Number of significant bits in each operand: 8, 16, 32, or 0
       (meaning, no operand).
     */
    byte opsize[2];
   
    /* 16-bit (or 8-bit with sign extension) displacement.
       only used with "mem" operand type.
     */
    word disp;

    /* Number of significant bits in disp: 8, 16, or 0. If 8, the disp
       variable will still be sign-extended to 16 bits.

       If this is zero, then there was no displacement encoded in the
       instruction. Note that this might be nonzero and disp might still
       be zero, in which case a displacement of zero was encoded in the
       instruction. This is relevant for the [bp+disp] addressing mode
       in particular: since the 8086 does not have a [bp] addressing
       mode, memory references that use the bp register will always
       have a displacement encoded in the instruction, even if that
       displacement is zero.
     */
    byte dispsize;
};

/*
Decode a sequence of bytes into a struct a86d_ins, representing a single
8086 instruction.

- ins: Pointer to a struct a86d_ins. The struct will be completely
overwritten.

- next_byte: Function that stores the next byte of machine code to the
pointer specified in its first argument. next_byte must return zero if
it provided the next byte, or nonzero to signal the decoder to stop
processing (for example, if there are no more bytes left in the input).

- data: Arbitrary data that gets passed in as the second argument to
next_byte each time 86d calls that function. It's up to the caller to
decide if/how to make use of this. NULL is okay here--the 86d library
will never dereference this pointer or do anything with it other than
pass it to next_byte. You might like to use this for tracking some
state, such as a pointer into an input stream, or the raw bytes that
have been processed.

The library will call the given next_byte function each time it needs
another byte of input, until a complete instruction has been fetched or
until next_byte returns nonzero. If the input is not a valid instruction
then the byte sequence will decode to an "a86_op_db" type of
instruction.

The return value is the number of times the next_byte function was
called, minus one. This is the same as the ins->length structure member.

This API allows for much greater flexibility than passing in an array of
bytes would. For an example, refer to main.c, which pulls the bytes
directly from a stdio file stream using fgetc.
*/
size_t a86d_decode(struct a86d_ins* ins, int (*next_byte)(byte*, void*),
                   void* data);

/*
Convert a struct a86d_ins to a nul-terminated string of assembly
language, and copy that line to the provided outbuf. The outbuf shall be
at least A86D_DISASM_OUTBUF_LEN chars long. The caller shall already
have fully initialized the a86d_ins (typically by calling a86d_decode)
before passing it into this function. The return value is outbuf.

For the special pseudo-instruction type "a86d_op_db", this function will
output only the string "db", since the struct a86d_ins does not contain
the actual raw bytes. If you want the actual bytes to follow the "db",
you'll have to collect them yourself.
*/
char* a86d_disasm(struct a86d_ins* ins, char* outbuf);

/*
Convert relative address operands to absolute values. For each operand
of type a86d_oper_relimm in the specified struct a86d_ins, this function
will add the provided offset to that operand. This is meant to be used
on relative calls and jumps so that you can display the absolute
address, so typically you will want to provide the absolute address of
the next instruction as the offset parameter to this function. You'll
want to call this before calling a86d_disasm. See main.c for an example.
*/
void a86d_rel_to_abs(struct a86d_ins* ins, word offset);

#endif /* A86D_H */
