/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bits.h"
#include "cpu.h"
#include "pic.h"
#include "clock.h"

/* TODO separate out diskette.h later. */
#include "diskette.h"
#include "memory.h"

#include "86d.h"

#define CF 1
#define PF 4
#define AF 16
#define ZF 64
#define SF 128
#define TF 256
#define IF 512
#define DF 1024
#define OF 2048
#define ax (AX.w)
#define al (AX.b[0])
#define ah (AX.b[1])
#define bx (BX.w)
#define bl (BX.b[0])
#define bh (BX.b[1])
#define cx (CX.w)
#define cl (CX.b[0])
#define ch (CX.b[1])
#define dx (DX.w)
#define dl (DX.b[0])
#define dh (DX.b[1])

#define getf(f) ((flags & (f)) != 0)

static int parity(byte b) {
    int n = b & 1;
    n += (b & 2) >> 1;
    n += (b & 4) >> 2;
    n += (b & 8) >> 3;
    n += (b & 16) >> 4;
    n += (b & 32) >> 5;
    n += (b & 64) >> 6;
    n += (b & 128) >> 7;
    return !(n & 1);
}

static dword sign_extend32(word in) {
    dword tmp = in;
    if (in & 0x8000) {
        tmp |= 0xffff0000;
    } else {
        tmp &=~ 0xffff0000;
    }
    return tmp;
}

static word sign_extend16(byte in) {
    word tmp = in;
    if (in & 0x80) {
        tmp |= 0xff00;
    } else {
        tmp &=~ 0xff00;
    }
    return tmp;
}


static struct a86d_ins current_ins;
static byte halted = 0;
static byte int0_fired = 0;

static dword ONE = 1;

struct location {
    enum a86d_oper_type type;
    byte size;
    dword data; /* reg, imm, or seg:off */
};


static union {
    byte b[2];
    word w;
} AX, BX, CX, DX;

word bp, sp, di, si, flags, cs, ds, es, ss, ip;

#define LB(n) ((byte)((n) & 0xff))
#define HB(n) ((byte)(((n) & 0xff00) >> 8))
#define LW(n) ((word)((n) & 0xffff))
#define HW(n) ((word)(((n) & 0xffff0000) >> 16))

static void setf(int f, int cond) {
    if (cond) {
        flags |= f;
    } else {
        flags &=~ f;
    }
}

static void set_szp16(word src) {
    setf(SF, src & 0x8000);
    setf(ZF, !src);
    setf(PF, parity(LB(src)));
}

static void set_szp8(byte src) {
    setf(SF, src & 0x80);
    setf(ZF, !src);
    setf(PF, parity(src));
}

int decoder_next_byte(byte* out, void* unused) {
    (void)unused;
    *out = load8(cs, ip++);
    return 0;
}

static inline void PUSH(word src) {
  sp -= 2;
  store16(ss, sp, src);
}

static inline void POP(word* dest) {
  *dest = load16(ss, sp);
  sp += 2;
}

dword io_in_dummy_func(word unused) {
    (void)unused;
    return UINT32_MAX;
}

void io_out_dummy_func(word unused1, word unused2) {
    (void)unused1, (void)unused2;
}

static dword (*io_in_func)(word) = io_in_dummy_func;
static void (*io_out_func)(word, word) = io_out_dummy_func;

void cpu_set_io_in_func(dword (*func)(word)) {
    io_in_func = func;
}

void cpu_set_io_out_func(void (*func)(word, word)) {
    io_out_func = func;
}

void cpu_reset(void) {
    ax = 0;
    bx = 0;
    cx = 0;
    dx = 0;
    bp = 0;
    sp = 0;
    di = 0;
    si = 0;
    /* top nibble is always set high on 8086/8088. Not true on later CPUs. */
    flags |= 0xf002;
    /* set interrupt flag to start? */
    /* setf(IF, 1); */
    cs = 0xffff;/* at the 286 (? 186?) this start point changes to f000:fff0 */
    ip = 0;
    ds = 0;
    es = 0;
    ss = 0;
    halted = 0;
    int0_fired = 0;
}

static void hlt(void) {
    halted = 1;
}

/* for debugging use only. not part of emulation */
#define die(s) (puts(s), hlt(), abort())

word segval(byte seg) {
    switch (seg) {
    case a86d_seg_es: return es;
    case a86d_seg_cs: return cs;
    case a86d_seg_ss: return ss;
    case a86d_seg_ds: return ds;
    case a86d_seg_none: return ds; /* bp addressing mode does not use this. */
    default: die("invalid seg enum");
    }
}

void set_szp(dword* dest, byte destsize) {
    if (destsize == 8) {
        set_szp8(LB(*dest));
    } else if (destsize == 16) {
        set_szp16(LW(*dest));
    } else {
        abort();
    }
}

void set_cf(dword* dest, byte destsize) {
    if (destsize == 8) {
        setf(CF, *dest > 0xff);
    } else if (destsize == 16) {
        setf(CF, *dest > 0xffff);
    } else {
        abort();
    }
}

static void add(dword* dest, dword* src, byte destsize) {
    dword olddest = *dest;
    *dest += *src;
    setf(AF, (*dest & 15) + (*src & 15) > 15);
    set_cf(dest, destsize);
    set_szp(dest, destsize);
    setf(OF, ((*dest ^ *src) & (*dest ^ olddest)) & (1 << (destsize - 1)));
}

static void adc(dword* dest, dword* src, byte destsize) {
    dword olddest = *dest;
    *dest += *src;
    *dest += getf(CF);
    setf(AF, (*dest & 15) + (*src & 15) > 15);
    set_cf(dest, destsize);
    set_szp(dest, destsize);
    setf(OF, ((*dest ^ *src) & (*dest ^ olddest)) & (1 << (destsize - 1)));
}

static void inc(dword* dest, byte destsize) {
    int tmpcf = getf(CF);
    add(dest, &ONE, destsize);
    setf(CF, tmpcf);
}

static void sub(dword* dest, dword* src, byte destsize) {
    dword olddest = *dest;
    *dest -= *src;
    setf(AF, (*dest & 15) + (*src & 15) > 15);
    set_cf(dest, destsize);
    set_szp(dest, destsize);
    setf(OF, ((olddest ^ *src) & (*dest ^ olddest)) & (1 << (destsize - 1)));
}

static void sbb(dword* dest, dword* src, byte destsize) {
    dword olddest = *dest;
    *dest -= *src;
    *dest -= getf(CF);
    setf(AF, (*dest & 15) + (*src & 15) > 15);
    set_cf(dest, destsize);
    set_szp(dest, destsize);
    setf(OF, ((olddest ^ *src) & (*dest ^ olddest)) & (1 << (destsize - 1)));
}

static void dec(dword* dest, byte destsize) {
    int tmpcf = getf(CF);
    sub(dest, &ONE, destsize);
    setf(CF, tmpcf);
}

static void neg(dword* dest, byte destsize) {
    dword src;
    switch (destsize) {
    case 8:
        if (LB(*dest) == 0x80) {
            setf(OF, 1);
            return;
        }
        break;
    case 16:
        if (LW(*dest) == 0x8000) {
            setf(OF, 1);
            return;
        }
        break;
    default: abort();
    }
    src = *dest;
    *dest = 0;
    sub(dest, &src, destsize);
}

static void mul(dword* dest, byte destsize) {
    dword tmpd = 0;
    word tmpw = 0;
    if (destsize == 16) {
        tmpd = *dest;
        tmpd *= ax;
        dx = HW(tmpd);
        ax = LW(tmpd);
        setf(CF, dx);
        setf(OF, dx);
    } else if (destsize == 8) {
        tmpw = *dest;
        tmpw *= al;
        ax = tmpw;
        setf(CF, ah);
        setf(OF, ah);
    } else {
        abort();
    }
}

static void imul(dword* dest, byte destsize) {
    dword tmpd = 0;
    word tmpw = 0;
    word tmpal;
    dword tmpax;
    if (destsize == 16) {
        tmpd = sign_extend32(*dest);
        tmpax = sign_extend32(ax);
        tmpd *= tmpax;
        dx = HW(tmpd);
        ax = LW(tmpd);
        setf(CF, dx != 0xffff && dx);
        setf(OF, dx != 0xffff && dx);
    } else if (destsize == 8) {
        tmpw = sign_extend16(*dest);
        tmpal = sign_extend16(al);
        tmpw *= tmpal;
        al = LB(tmpw);
        ah = HB(tmpw);
        setf(CF, ah != 0xff && ah);
        setf(OF, ah != 0xff && ah);
    } else {
        abort();
    }
}

static void aaa(void) {
    if (getf(AF) || ((al & 0xf) >= 10)) {
        al += 6;
        ah += 1;
        setf(AF, 1);
    }
    al &= 0xf;
    setf(CF, getf(AF));
}

static void aad(byte scale) {
    ah *= scale;
    al += ah;
    ah = 0;
    set_szp8(al);
}

void diskette_io(void) {
    int result;
    switch (ah) {
    case 0:
        result = diskette_reset(al);
        break;
    case 1:
        /* status stub */
        result = 0;
        al = 0;
        break;
    case 2:
        result = diskette_read(ch, dh, cl, al, dl, es, bx);
        if (result >= 0) {
            al = result;
        }
        break;
    case 3:
        result = diskette_write(ch, dh, cl, al, dl, es, bx);
        if (result >= 0) {
            al = result;
        }
        break;
    case 4:
        result = diskette_verify(ch, dh, cl, al, dl);
        if (result >= 0) {
            al = result;
        }
        break;
    case 5:
        result = diskette_format(ch, dh, cl, al, dl, es, bx);
        break;
    default:
        /* invalid */
        ah = 1;
        al = 0;
        setf(CF, 1);
        return;
    }
    setf(CF, result < 0);
}

static void int_(byte num) {
    /* TODO figure out why keyboard interrupt is not firing at the correct
       time during BIOS POST. Step through BIOS to investigate.
    */
    if (num == 0x13) {
        diskette_io();
        return;
    }
    PUSH(flags);
    setf(IF, 0);
    setf(TF, 0);
    PUSH(cs);
    cs = load16(0, 2 + 4 * num);
    PUSH(ip);
    ip = load16(0, 4 * num);
}


static void aam(byte scale) {
    if (!scale) {
        int0_fired = 1;
        return;
    }
    byte tmp = al / scale;
    ah = tmp;
    al = al % 0x0a;
    set_szp8(al);
}

static void aas(void) {
    if (getf(AF) || ((al & 0xf) >= 10)) {
        al -= 6;
        ah -= 1;
        setf(AF, 1);
    }
    al &= 0xf;
    setf(CF, getf(AF));
}

static void daa(void) {
    if (getf(AF) || ((al & 0xf) >= 10)) {
        al += 6;
        setf(AF, 1);
    }
    if (getf(CF) || (((al & 0xf0) >> 4) >= 10)) {
        al += 0x60;
        setf(CF, 1);
    }
}

static void das(void) {
    if (getf(AF) || ((al & 0xf) >= 10)) {
        al -= 6;
        setf(AF, 1);
    }
    if (getf(CF) || (((al & 0xf0) >> 4) >= 10)) {
        al -= 0x60;
        setf(CF, 1);
    }
}


static void in(dword* dest, dword* port, byte destsize) {
    dword received;
    received = io_in_func(LW(*port));
    if (received == UINT32_MAX) {
        /* not implemented */
        return;
    }
    if (destsize == 8) {
        *dest = LB(received);
    } else if (destsize == 16) {
        *dest = LW(received);
    } else {
        abort();
    }
}

static void out(dword* port, dword* src, byte destsize) {
    if (destsize == 8) {
        io_out_func(LW(*port), (word)LB(*src));
    } else if (destsize == 16) {
        io_out_func(LW(*port), LW(*src));
    } else {
        abort();
    }
}

static void into(void) {
    if (getf(OF)) {
        int_(4);
    }
}
static void div16(word src) {
    dword dividend = (dx << 16) + ax;
    dword rem = 0;
    dword quot = 0;
    if (!src) {
        int0_fired = 1;
        return;
    }
    quot = dividend / src;
    rem = dividend % src;
    if (quot > 0xffff) {
        int0_fired = 1;
        return;
    }
    ax = LW(quot);
    dx = LW(rem);
}

static void div8(byte src) {
    word dividend = (ah << 8) + al;
    word rem = 0;
    word quot = 0;
    if (!src) {
        int0_fired = 1;
        return;
    }
    quot = dividend / src;
    rem = dividend % src;
    if (quot > 0xff) {
        int0_fired = 1;
        return;
    }
    al = LB(quot);
    ah = LB(rem);
}

static void div_(dword* src, byte destsize) {
    if (destsize == 16) {
        div16(LW(*src));
    } else if (destsize == 8) {
        div8(LB(*src));
    } else {
        abort();
    }
}

static void idiv16(word src) {
    dword dx32 = sign_extend32(dx) << 16;
    dword ax32 = ax;
    dword dividend = dx32 | ax32;
    dword divisor = sign_extend32(src);
    dword rem = 0;
    dword quot = 0;
    byte src_neg = (divisor & 0x80000000) != 0;
    byte dest_neg = (dividend & 0x80000000) != 0;
    if (!src) {
        int0_fired = 1;
        return;
    }
    /* sign flip */
    if (src_neg) {
        divisor = ~divisor + 1;
    }
    if (dest_neg) {
        dividend = ~dividend + 1;
    }
    quot = dividend / divisor;
    rem = dividend % divisor;
    if (HW(quot) || quot == 0x8000) {
        int0_fired = 1;
        return;
    }
    ax = LW(quot);
    dx = LW(rem);
    if ((src_neg || dest_neg) && !(src_neg && dest_neg)) {
        /* negative sign restore */
        ax = ~ax + 1;
        dx = ~dx + 1;
    }
}

static void idiv8(byte src) {
    word dividend = ax;
    word divisor = sign_extend16(src);
    word rem = 0;
    word quot = 0;
    byte src_neg = (divisor & 0x8000) != 0;
    byte dest_neg = (dividend & 0x8000) != 0;
    if (!src) {
        int0_fired = 1;
        return;
    }
    /* sign flip */
    if (src_neg) {
        divisor = ~divisor + 1;
    }
    if (dest_neg) {
        dividend = ~dividend + 1;
    }
    quot = dividend / divisor;
    rem = dividend % divisor;
    if (HB(quot) || quot == 0x80) {
        int0_fired = 1;
        return;
    }
    al = LB(quot);
    ah = LB(rem);
    if ((src_neg || dest_neg) && !(src_neg && dest_neg)) {
        /* negative sign restore */
        al = ~al + 1;
        ah = ~ah + 1;
    }
}

static void idiv(dword* src, byte destsize) {
    if (destsize == 16) {
        idiv16(LW(*src));
    } else if (destsize == 8) {
        idiv8(LB(*src));
    } else {
        abort();
    }
}

static void not(dword *dest, byte destsize) {
    (void)destsize; /* don't care */
    *dest = ~*dest;
}

static void and(dword *dest, dword* src, byte destsize) {
    *dest &= *src;
    setf(OF, 0);
    setf(CF, 0);
    set_szp(dest, destsize);
}

static void or(dword *dest, dword* src, byte destsize) {
    *dest |= *src;
    setf(OF, 0);
    setf(CF, 0);
    set_szp(dest, destsize);
}

static void xor(dword *dest, dword* src, byte destsize) {
    *dest ^= *src;
    setf(OF, 0);
    setf(CF, 0);
    set_szp(dest, destsize);
}

static void shl16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    while (amount) {
        setf(CF, (*dest) & (1<<15));
        *dest <<= 1;
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void shl8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    while (amount) {
        setf(CF, (*dest) & (1<<7));
        *dest <<= 1;
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void shl(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        shl16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        shl8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}

static void shr16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    while (amount) {
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void shr8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    while (amount) {
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void shr(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        shr16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        shr8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}

static void sar16(word* dest, byte amount) {
    while (amount) {
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        if (*dest & (1 << 14)) {
            *dest |= (1 << 15);
        }
        amount--;
    }
    setf(OF, 0);
    set_szp16(*dest);
}

static void sar8(byte* dest, byte amount) {
    while (amount) {
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        if (*dest & (1 << 6)) {
            *dest |= (1 << 7);
        }
        amount--;
    }
    setf(OF, 0);
    set_szp8(*dest);
}

static void sar(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        sar16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        sar8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}

static void rol16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    byte high;
    while (amount) {
        high = is_sign16(*dest);
        setf(CF, (*dest) & (1<<15));
        *dest <<= 1;
        *dest |= high;
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void rol8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    byte high;
    while (amount) {
        high = is_sign8(*dest);
        setf(CF, (*dest) & (1<<7));
        *dest <<= 1;
        *dest |= high;
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void rol(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        rol16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        rol8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}

static void ror16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    byte low;
    while (amount) {
        low = *dest & 1;
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        *dest |= low << 15;
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void ror8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    byte low;
    while (amount) {
        low = *dest & 1;
        setf(CF, (*dest) & 1);
        *dest >>= 1;
        *dest |= low << 7;
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void ror(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        ror16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        ror8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}


static void rcl16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    byte newcf;
    while (amount) {
        newcf = is_sign16(*dest);
        *dest <<= 1;
        if (getf(CF)) {
            *dest |= 1;
        } else {
            *dest &=~ 1;
        }
        setf(CF, newcf);
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void rcl8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    byte newcf;
    while (amount) {
        newcf = is_sign8(*dest);
        *dest <<= 1;
        if (getf(CF)) {
            *dest |= 1;
        } else {
            *dest &=~ 1;
        }
        setf(CF, newcf);
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void rcl(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        rcl16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        rcl8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}


static void rcr16(word* dest, byte amount) {
    byte oldsign = is_sign16(*dest);
    byte newcf;
    while (amount) {
        newcf = *dest & 1;
        *dest >>= 1;
        if (getf(CF)) {
            *dest |= 1 << 15;
        } else {
            *dest &=~ (1 << 15);
        }
        setf(CF, newcf);
        amount--;
    }
    setf(OF, oldsign ^ is_sign16(*dest));
    set_szp16(*dest);
}

static void rcr8(byte* dest, byte amount) {
    byte oldsign = is_sign8(*dest);
    byte newcf;
    while (amount) {
        newcf = *dest & 1;
        *dest >>= 1;
        if (getf(CF)) {
            *dest |= (1 << 7);
        } else {
            *dest &=~ (1 << 7);
        }
        setf(CF, newcf);
        amount--;
    }
    setf(OF, oldsign ^ is_sign8(*dest));
    set_szp8(*dest);
}

static void rcr(dword* src, dword* dest, byte destsize) {
    byte b;
    word w;
    if (destsize == 16) {
        w = LW(*src);
        rcr16(&w, LB(*dest));
        *src = w;
    } else if (destsize == 8) {
        b = LB(*src);
        rcr8(&b, LB(*dest));
        *src = b;
    } else {
        abort();
    }
}


static void movsb(word seg) {
    store8(es, di, load8(seg, si));
    if (getf(DF)) {
        si--; di--;
    } else {
        si++; di++;
    }
}

static void movsw(word seg) {
    store16(es, di, load16(seg, si));
    if (getf(DF)) {
        si -= 2; di -= 2;
    } else {
        si += 2; di += 2;
    }
}

static void cmpsb(word seg) {
    dword b1 = load8(seg, si);
    dword b2 = load8(es, di);
    sub(&b1, &b2, 8);
    if (getf(DF)) {
        si--; di--;
    } else {
        si++; di++;
    }
}

static void cmpsw(word seg) {
    dword w1 = load16(seg, si);
    dword w2 = load16(es, di);
    sub(&w1, &w2, 16);
    if (getf(DF)) {
        si -= 2; di -= 2;
    } else {
        si += 2; di += 2;
    }
}

static void scasb(void) {
    dword b1 = al;
    dword b2 = load8(es, di);
    sub(&b1, &b2, 8);
    if (getf(DF)) {
        di--;
    } else {
        di++;
    }
}

static void scasw(void) {
    dword w1 = ax;
    dword w2 = load16(es, di);
    sub(&w1, &w2, 16);
    if (getf(DF)) {
        di -= 2;
    } else {
        di += 2;
    }
}

static void lodsb(word seg) {
    al = load8(seg, si);
    if (getf(DF)) {
        si--;
    } else {
        si++;
    }
}

static void lodsw(word seg) {
    ax = load16(seg, si);
    if (getf(DF)) {
        si -= 2;
    } else {
        si += 2;
    }
}

static void stosb(void) {
    store8(es, di, al);
    if (getf(DF)) {
        di--;
    } else {
        di++;
    }
}

static void stosw(void) {
    store16(es, di, ax);
    if (getf(DF)) {
        di -= 2;
    } else {
        di += 2;
    }
}

void process_interrupts(void) {
    byte irqs_raised = 0;
    if (getf(IF)) {
        irqs_raised = pic_poll();
        if (irqs_raised & 1) {
            int_(8);
        }
        if (irqs_raised & 2) {
            int_(9);
        }
    }
    /* single step trap */
    if (getf(TF)) {
        int_(1);
    }
}

static void string_inst(struct a86d_ins* ins) {
    enum a86d_op op = ins->op;
    enum a86d_rep_mode rep_mode = ins->rep_mode;
    word seg = segval(ins->seg);
rep_loop:
    if (rep_mode != a86d_no_rep) {
        if (!cx) {
            return;
        }
        cx--;
    }
    switch (op) {
    case a86d_op_movsb: movsb(seg); break;
    case a86d_op_movsw: movsw(seg); break;
    case a86d_op_cmpsb: cmpsb(seg); break;
    case a86d_op_cmpsw: cmpsw(seg); break;
    case a86d_op_stosb: stosb(); break;
    case a86d_op_stosw: stosw(); break;
    case a86d_op_lodsb: lodsb(seg); break;
    case a86d_op_lodsw: lodsw(seg); break;
    case a86d_op_scasb: scasb(); break;
    case a86d_op_scasw: scasw(); break;
    default:   die("invalid string inst"); break;
    }
    if (rep_mode != a86d_no_rep
        && (op == a86d_op_scasb
            || op == a86d_op_scasw
            || op == a86d_op_cmpsb
            || op == a86d_op_cmpsw)) {
        if (rep_mode == a86d_repne) {
            if (getf(ZF)) {
                return;
            }
        } else { /* repe */
            if (!getf(ZF)) {
                return;
            }
        }
    }
    if (rep_mode != a86d_no_rep) {
        /* TODO more specific timing */
        clock_advance(2);
        process_interrupts();
        goto rep_loop;
    }
}

static void mov(dword* dest, dword* src, byte unused) {
    (void)unused;
    *dest = *src;
}

static void xchg(dword* dest, dword* src, byte unused) {
    (void)unused;
    dword tmp = *src;
    *src = *dest;
    *dest = tmp;
}

static void xlat(word seg) {
    al = load8(seg, bx + al);
}

static void lahf(void) {
    /* reserved bits are included */
    ah = LB(flags);
}

static void sahf(void) {
    setf(SF, (ah & SF));
    setf(ZF, (ah & ZF));
    setf(AF, (ah & AF));
    setf(PF, (ah & PF));
    setf(CF, (ah & CF));
}

static void call_far(dword* dest, byte destsize) {
    (void)destsize; /* always 32 */
    PUSH(cs);
    PUSH(ip);
    ip = LW(*dest);
    cs = HW(*dest);
}

static void call_near(dword* dest,  byte destsize) {
    (void)destsize; /* 16 */
    PUSH(ip);
    ip = LW(*dest);
}

static void call_disp(dword* dest, byte destsize) {
    (void)destsize; /* 16 */
    PUSH(ip);
    ip += LW(*dest);
}

static void jmp_far(dword* dest, byte destsize) {
    (void)destsize; /* 32 */
    ip = LW(*dest);
    cs = HW(*dest);
}

static void jmp_near(dword* dest, byte destsize) {
    (void)destsize; /* 16 */
    ip = LW(*dest);
}

static void jmp_disp(dword* dest, byte destsize) {
    if (destsize == 8) {
        ip += sign_extend16(LB(*dest));
    } else {
        ip += LW(*dest);
    }
}

static void jmp_disp8(int8_t disp) {
    ip += disp;
}

static void loop(int8_t disp) {
    cx -= 1;
    if (cx) {
        jmp_disp8(disp);
    }
}

static void loope(int8_t disp) {
    cx -= 1;
    if (cx && getf(ZF)) {
        jmp_disp8(disp);
    }
}
#define loopz loope

static void loopne(int8_t disp) {
    cx -= 1;
    if (cx && !getf(ZF)) {
        jmp_disp8(disp);
    }
}
#define loopnz loopne

static void ja(int8_t disp) {
    if (!getf(ZF) && !getf(CF)) {
        jmp_disp8(disp);
    }
}
#define jnbe ja

static void jae(int8_t disp) {
    if (!getf(CF)) {
        jmp_disp8(disp);
    }
}
#define jnb jae
#define jnc jae

static void jb(int8_t disp) {
    if (getf(CF)) {
        jmp_disp8(disp);
    }
}
#define jnae jb
#define jc jb

static void jna(int8_t disp) {
    if (getf(ZF) || getf(CF)) {
        jmp_disp8(disp);
    }
}
#define jbe jna

static void jg(int8_t disp) {
    if (!getf(ZF) && getf(SF) == getf(OF)) {
        jmp_disp8(disp);
    }
}

#define jnle jg

static void jl(int8_t disp) {
    if (getf(SF) != getf(OF)) {
        jmp_disp8(disp);
    }
}
#define jnge jl

static void jle(int8_t disp) {
    if (getf(ZF) || getf(SF) != getf(OF)) {
        jmp_disp8(disp);
    }
}
#define jng jle

static void jge(int8_t disp) {
    if (getf(SF) == getf(OF)) {
        jmp_disp8(disp);
    }
}
#define jnl jge

static void jp(int8_t disp) {
    if (getf(PF)) {
        jmp_disp8(disp);
    }
}
#define jpe jp

static void jnp(int8_t disp) {
    if (!getf(PF)) {
        jmp_disp8(disp);
    }
}
#define jpo jnp

static void js(int8_t disp) {
    if (getf(SF)) {
        jmp_disp8(disp);
    }
}

static void jns(int8_t disp) {
    if (!getf(SF)) {
        jmp_disp8(disp);
    }
}

static void jcxz(int8_t disp) {
    if (!cx) {
        jmp_disp8(disp);
    }
}

static void jo(int8_t disp) {
    if (getf(OF)) {
        jmp_disp8(disp);
    }
}

static void jno(int8_t disp) {
    if (!getf(OF)) {
        jmp_disp8(disp);
    }
}

static void jz(int8_t disp) {
    if (getf(ZF)) {
        jmp_disp8(disp);
    }
}
#define je jz

static void jnz(int8_t disp) {
    if (!getf(ZF)) {
        jmp_disp8(disp);
    }
}
#define jne jnz

static void ret(dword* popval, byte destsize) {
    (void)destsize;
    POP(&ip);
    sp += LW(*popval);
}

static void retf(dword* popval, byte destsize) {
    (void)destsize;
    POP(&ip);
    POP(&cs);
    sp += LW(*popval);
}

static void iret(void) {
    POP(&ip);
    POP(&cs);
    POP(&flags);
}

static void cmc(void) {
    if (getf(CF)) {
        setf(CF, 0);
    } else {
        setf(CF, 1);
    }
}

static void salc(void) {
    if (getf(CF)) {
        al = 0xff;
    } else {
        al = 0;
    }
}

static void lds(dword* dest, dword* src, byte destsize) {
    (void)destsize; /* 16 bits */
    *dest = LW(*src);
    ds = HW(*src);
}

static void les(dword* dest, dword* src, byte destsize) {
    (void)destsize; /* 16 bits */
    *dest = LW(*src);
    es = HW(*src);
}

static void nop(void) {
    /* nothing */
}

void pop(dword* dest, byte destsize) {
    (void)destsize;
    *dest = load16(ss, sp);
    sp += 2;
}

void push(dword* dest, byte destsize) {
    (void)destsize;
    sp -= 2;
    store16(ss, sp, LW(*dest));
}

static word* decode_reg16(byte bits3) {
    switch (bits3 & 0x7) {
    case 0: return &ax;
    case 1: return &cx;
    case 2: return &dx;
    case 3: return &bx;
    case 4: return &sp;
    case 5: return &bp;
    case 6: return &si;
    case 7: return &di;
    default: abort();
    }
}

static word* decode_segreg(byte bits2) {
    switch (bits2 & 0x3) {
    case 0: return &es;
    case 1: return &cs;
    case 2: return &ss;
    case 3: return &ds;
    default: abort();
    }
}

static byte* decode_reg8(byte bits3) {
    switch (bits3 & 0x7) {
    case 0: return &al;
    case 1: return &cl;
    case 2: return &dl;
    case 3: return &bl;
    case 4: return &ah;
    case 5: return &ch;
    case 6: return &dh;
    case 7: return &bh;
    default: abort();
    }
}

void make_loc(struct location* loc, struct a86d_ins* ins, byte i) {
    loc->size = ins->opsize[i];
    loc->type = ins->optype[i];
    switch (ins->optype[i]) {
    case a86d_oper_none:
    case a86d_oper_reg:
    case a86d_oper_segreg:
    case a86d_oper_imm:
    case a86d_oper_relimm:
        loc->data = ins->operand[i];
        break;
    case a86d_oper_directaddr:
        /* data is a 16 bit offset. size refers to size of memory pointed to */
        loc->data = LW(ins->operand[i]);
        loc->data &= ~0xffff0000;
        loc->data |= (segval(ins->seg) << 16);
        break;
    case a86d_oper_mem:
        switch (ins->operand[i] & 0x7) {
        case 0: loc->data = LW(bx + si + ins->disp); break;
        case 1: loc->data = LW(bx + di + ins->disp); break;
        case 2: loc->data = LW(bp + si + ins->disp); break;
        case 3: loc->data = LW(bp + di + ins->disp); break;
        case 4: loc->data = LW(00 + si + ins->disp); break;
        case 5: loc->data = LW(00 + di + ins->disp); break;
        case 6: loc->data = LW(00 + bp + ins->disp); break;
        case 7: loc->data = LW(00 + bx + ins->disp); break;
        default: abort();
        }
        if (ins->seg == a86d_seg_none &&
            (ins->operand[i] & 0x7) == 6) {
            /* bp uses ss by default */
            loc->data |= (ss << 16);
        } else {
            loc->data |= (segval(ins->seg) << 16);
        }
        break;
    default: abort();
    }
}

dword loc_fetch(struct location* loc) {
    word seg, off;
    switch (loc->type) {
    case a86d_oper_none:
        return 0;
    case a86d_oper_reg:
        switch (loc->size) {
        case 8: return *decode_reg8(LB(loc->data));
        case 16: return *decode_reg16(LB(loc->data));
        default: die("invalid register size");
        }
        break;
    case a86d_oper_segreg:
        return *decode_segreg(LB(loc->data));
    case a86d_oper_directaddr:
    case a86d_oper_mem:
        seg = HW(loc->data);
        off = LW(loc->data);
        switch (loc->size) {
        case 8: return load8(seg, off);
        case 16: return load16(seg, off);
        case 32: return load32(seg, off);
        default: die("invalid memory access size");
        }
        break;
    case a86d_oper_imm:
    case a86d_oper_relimm:
        switch (loc->size) {
        case 8:
        case 16:
        case 32:
            return loc->data;
        default: die("invalid immediate size");
        }
    default: abort();
    }
}

void loc_store(struct location* loc, dword val) {
    word seg;
    word off;
    switch (loc->type) {
    case a86d_oper_reg:
        switch (loc->size) {
        case 8: *decode_reg8(LB(loc->data)) = LB(val); break;
        case 16: *decode_reg16(LB(loc->data)) = LW(val); break;
        default: die("invalid register size");
        }
        break;
    case a86d_oper_segreg:
        *decode_segreg(LB(loc->data)) = LW(val);
        break;
    case a86d_oper_directaddr:
    case a86d_oper_mem:
        seg = HW(loc->data);
        off = LW(loc->data);
        switch (loc->size) {
        case 8: store8(seg, off, LB(val)); break;
        case 16: store16(seg, off, LW(val)); break;
        case 32: store32(seg, off, val); break;
        default: die("invalid memory access size");
        }
        break;
    case a86d_oper_imm:
    case a86d_oper_relimm:
    case a86d_oper_none:
        /* nothing */
        break;
    default: abort();
    }
}

static void lea(struct a86d_ins* ins) {
    struct location destloc = {0};
    struct location srcloc = {0};
    make_loc(&destloc, ins, 0);
    make_loc(&srcloc, ins, 1);
    loc_store(&destloc, LW(srcloc.data));
}

void do_ins_2_general(struct a86d_ins* ins,
            void (*func)(dword*, dword*, byte),
            int do_store_dest,
            int do_store_src) {
    struct location destloc = {0};
    struct location srcloc = {0};
    dword dest;
    dword src;
    make_loc(&destloc, ins, 0);
    make_loc(&srcloc, ins, 1);
    dest = loc_fetch(&destloc);
    src = loc_fetch(&srcloc);
    func(&dest, &src, destloc.size);
    if (int0_fired) {
        int0_fired = 0;
        int_(0);
        return;
    }
    if (do_store_dest) {
        loc_store(&destloc, dest);
    }
    if (do_store_src) {
        loc_store(&srcloc, src);
    }
}

void do_ins_1_general(struct a86d_ins* ins,
            void (*func)(dword*, byte),
            int do_store) {
    struct location destloc = {0};
    dword dest;
    make_loc(&destloc, ins, 0);
    dest = loc_fetch(&destloc);
    func(&dest, destloc.size);
    if (int0_fired) {
        int0_fired = 0;
        int_(0);
        return;
    }
    if (do_store) {
        loc_store(&destloc, dest);
    }
}

/* two operands, both modified */
void do_ins_wr(struct a86d_ins* ins,
            void (*func)(dword*, dword*, byte)) {
    do_ins_2_general(ins, func, 1, 0);
}

/* two operands, only first one modified */
void do_ins_ww(struct a86d_ins* ins,
            void (*func)(dword*, dword*, byte)) {
    do_ins_2_general(ins, func, 1, 1);
}

/* two operands, neither modified */
void do_ins_rr(struct a86d_ins* ins,
            void (*func)(dword*, dword*, byte)) {
    do_ins_2_general(ins, func, 0, 0);
}

/* one operand, modified */
void do_ins_w(struct a86d_ins* ins,
            void (*func)(dword*, byte)) {
    do_ins_1_general(ins, func, 1);
}

/* one operand, not modified */
void do_ins_r(struct a86d_ins* ins,
            void (*func)(dword*, byte)) {
    do_ins_1_general(ins, func, 0);
}

void do_wait(void) {
    /* TODO impl */
}

void pushf(void) {
    /* top nibble is always set high on 8086/8088. Not true on later CPUs. */
    flags |= 0xf002;
    PUSH(flags);
}

void popf(void) {
    flags |= 0xf002;
    POP(&flags);
    flags |= 0xf002;
}

void execute(struct a86d_ins* ins) {
    switch (ins->op) {
    case a86d_op_aaa: aaa(); break;
    case a86d_op_aad: aad((byte)ins->operand[0]); break;
    case a86d_op_aam: aam((byte)ins->operand[0]); break;
    case a86d_op_aas: aas(); break;
    case a86d_op_adc: do_ins_wr(ins, adc); break;
    case a86d_op_add: do_ins_wr(ins, add); break;
    case a86d_op_and: do_ins_wr(ins, and); break;
    case a86d_op_call_abs32: do_ins_r(ins, call_far); break;
    case a86d_op_call_mem16: do_ins_r(ins, call_near); break;
    case a86d_op_call_mem32: do_ins_r(ins, call_far); break;
    case a86d_op_call_rel16: do_ins_r(ins, call_disp); break;
    case a86d_op_cbw: ax = sign_extend16(al); break;
    case a86d_op_clc: setf(CF, 0); break;
    case a86d_op_cld: setf(DF, 0); break;
    case a86d_op_cli: setf(IF, 0); break;
    case a86d_op_cmc: cmc(); break;
    case a86d_op_cmp: do_ins_rr(ins, sub); break;
    case a86d_op_cmpsb: string_inst(ins); break;
    case a86d_op_cmpsw: string_inst(ins); break;
    case a86d_op_cwd: dx = HW(sign_extend32(ax)); break;
    case a86d_op_daa: daa(); break;
    case a86d_op_das: das(); break;
    case a86d_op_dec: do_ins_w(ins, dec); break;
    case a86d_op_div: do_ins_r(ins, div_); break;
    case a86d_op_esc: /* nothing */ break;
    case a86d_op_hlt: hlt(); break;
    case a86d_op_idiv: do_ins_r(ins, idiv); break;
    case a86d_op_imul: do_ins_r(ins, imul); break;
    case a86d_op_in: do_ins_wr(ins, in); break;
    case a86d_op_inc: do_ins_w(ins, inc); break;
    case a86d_op_int: int_((byte)ins->operand[0]); break;
    case a86d_op_int3: int_(3); break;
    case a86d_op_into: into(); break;
    case a86d_op_iret: iret(); break;
    case a86d_op_ja: ja((byte)ins->operand[0]); break;
    case a86d_op_jbe: jbe((byte)ins->operand[0]); break;
    case a86d_op_jc: jc((byte)ins->operand[0]); break;
    case a86d_op_jcxz: jcxz((byte)ins->operand[0]); break;
    case a86d_op_jz: jz((byte)ins->operand[0]); break;
    case a86d_op_jg: jg((byte)ins->operand[0]); break;
    case a86d_op_jge: jge((byte)ins->operand[0]); break;
    case a86d_op_jl: jl((byte)ins->operand[0]); break;
    case a86d_op_jle: jle((byte)ins->operand[0]); break;
    case a86d_op_jmp_abs32: do_ins_r(ins, jmp_far); break;
    case a86d_op_jmp_mem16: do_ins_r(ins, jmp_near); break;
    case a86d_op_jmp_mem32: do_ins_r(ins, jmp_far); break;
    case a86d_op_jmp_rel16: do_ins_r(ins, jmp_disp); break;
    case a86d_op_jmp_rel8: do_ins_r(ins, jmp_disp); break;
    case a86d_op_jnc: jnc((byte)ins->operand[0]); break;
    case a86d_op_jnz: jnz((byte)ins->operand[0]); break;
    case a86d_op_jno: jno((byte)ins->operand[0]); break;
    case a86d_op_jns: jns((byte)ins->operand[0]); break;
    case a86d_op_jo: jo((byte)ins->operand[0]); break;
    case a86d_op_jp: jp((byte)ins->operand[0]); break;
    case a86d_op_jpo: jpo((byte)ins->operand[0]); break;
    case a86d_op_js: js((byte)ins->operand[0]); break;
    case a86d_op_lahf: lahf(); break;
    case a86d_op_lds: do_ins_wr(ins, lds); break;
    case a86d_op_lea: lea(ins); break;
    case a86d_op_les: do_ins_wr(ins, les); break;
    case a86d_op_lodsb: string_inst(ins); break;
    case a86d_op_lodsw: string_inst(ins); break;
    case a86d_op_loop: loop((byte)ins->operand[0]); break;
    case a86d_op_loopnz: loopnz((byte)ins->operand[0]); break;
    case a86d_op_loopz: loopz((byte)ins->operand[0]); break;
    case a86d_op_mov: do_ins_wr(ins, mov); break;
    case a86d_op_movsb: string_inst(ins); break;
    case a86d_op_movsw: string_inst(ins); break;
    case a86d_op_mul: do_ins_r(ins, mul); break;
    case a86d_op_neg: do_ins_w(ins, neg); break;
    case a86d_op_nop: nop(); break;
    case a86d_op_not: do_ins_w(ins, not); break;
    case a86d_op_or: do_ins_wr(ins, or); break;
    case a86d_op_out: do_ins_rr(ins, out); break;
    case a86d_op_pop: do_ins_w(ins, pop); break;
    case a86d_op_popf: popf(); break;
    case a86d_op_push: do_ins_r(ins, push); break;
    case a86d_op_pushf: pushf(); break;
    case a86d_op_rcl: do_ins_wr(ins, rcl); break;
    case a86d_op_rcr: do_ins_wr(ins, rcr); break;
    case a86d_op_ret: do_ins_r(ins, ret); break;
    case a86d_op_retf: do_ins_r(ins, retf); break;
    case a86d_op_rol: do_ins_wr(ins, rol); break;
    case a86d_op_ror: do_ins_wr(ins, ror); break;
    case a86d_op_sahf: sahf(); break;
    case a86d_op_salc: salc(); break;
    case a86d_op_sar: do_ins_wr(ins, sar); break;
    case a86d_op_sbb: do_ins_wr(ins, sbb); break;
    case a86d_op_scasb: string_inst(ins); break;
    case a86d_op_scasw: string_inst(ins); break;
    case a86d_op_shl: do_ins_wr(ins, shl); break;
    case a86d_op_shr: do_ins_wr(ins, shr); break;
    case a86d_op_stc: setf(CF, 1); break;
    case a86d_op_std: setf(DF, 1); break;
    case a86d_op_sti: setf(IF, 1); break;
    case a86d_op_stosb: string_inst(ins); break;
    case a86d_op_stosw: string_inst(ins); break;
    case a86d_op_sub: do_ins_wr(ins, sub); break;
    case a86d_op_test: do_ins_rr(ins, and); break;
    case a86d_op_wait: do_wait(); break;
    case a86d_op_xchg: do_ins_ww(ins, xchg); break;
    case a86d_op_xlat: xlat(segval(ins->seg)); break;
    case a86d_op_xor: do_ins_wr(ins, xor); break;
    case a86d_op_none: abort();
    case a86d_op_db: break; /* nothing */
    default: abort();
    }
}

void cpu_fetch(void) {
    if (!halted) {
        a86d_decode(&current_ins, decoder_next_byte, NULL);
    }
}

void cpu_execute(void) {
    if (!halted) {
        execute(&current_ins);
    }
    /* TODO use real cycle counts for each instruction */
    clock_advance(2);
    process_interrupts();
}

void cpu_dump_state(void) {
    char outbuf[64] = {0};
    a86d_disasm(&current_ins, outbuf);
    puts("+--------------------------------------+-cs---ip---+-oditsz-a-p-c------+");
    printf("| %-36s | %04hx:%04hx | %d%d%d%d%d%d %d %d %d      |\n",
            outbuf, cs, ip,
            getf(OF), getf(DF), getf(IF), getf(TF), getf(SF),
            getf(ZF), getf(AF), getf(PF), getf(CF));
    puts("+--------------------------------------+-----------+-----------+-------+");
    puts("| ax   bx   cx   dx     ss   sp   bp     ds   si     es   di     flags |");
    printf("| %04hx %04hx %04hx %04hx   %04hx %04hx %04hx   %04hx %04hx   %04hx %04hx   %04hx  |\n",
           ax, bx, cx, dx, ss, sp, bp, ds, si, es, di, flags);
}

word cpu_get_ip(void) {
    return ip;
}
word cpu_get_cs(void) {
    return cs;
}

void cpu_set_cs(word w) {
    cs = w;
}
void cpu_set_ip(word w) {
    ip = w;
}

byte cpu_is_halted(void) {
    return halted;
}
