/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "86d.h"
#include "bits.h"
#include "breakpoint.h"
#include "cpu.h"
#include "diskette.h"
#include "dma.h"
#include "mda.h"
#include "memory.h"
#include "term.h"
#include "timer.h"
#include "periph.h"
#include "pic.h"
#include "str.h"
    
static int hextoul(const char* s, dword* out) {
    dword count = 0;
    if (!s) {
        return -1;
    }
    /* skip 0x */
    if (s[0] == '0' && (s[1] == 'x' || s[1] == 'X')) {
        s += 2;
    }
    /* trailing 'h' is allowed */
    for (*out = 0; *s && *s != 'h' && *s != 'H'; s++, count++) {
        *out <<= 4;
        if (*s >= '0' && *s <= '9') {
            *out += *s - '0';
        } else if (*s >= 'A' && *s <= 'F') {
            *out += *s - 'A' + 10;
        } else if (*s >= 'a' && *s <= 'f') {
            *out += *s - 'a' + 10;
        } else {
            return -1;
        }
    }
    /* max 8 digits */
    if (count >= 1 && count <= 8) {
        return 0;
    }
    return -1;
}

static int segoff_parse(char* s, word* seg_out, word* off_out, word default_cs) {
    char* split;
    dword addr;
    dword seg;
    dword off;
    /* strip leading spaces */
    while (*s && *s == ' ') {
        s++;
    }
    if (!s) {
        return -1;
    }
    split = s;
    while (*split && *split != ':') {
        split++;
    }
    if (*split == ':') {
        *split = '\0';
        if (hextoul(s, &seg) < 0 || seg > 0xffff) {
            return -1;
        }
        if (hextoul(split+1, &off) < 0 || off > 0xffff) {
            return -1;
        }
        *seg_out = seg & 0xffff;
        *off_out = off & 0xffff;
        return 0;
    } else {
        if (hextoul(s, &addr) < 0) {
            return -1;
        }
        if (addr > 0xfffff) {
            return -1;
        }
        if (addr < 0x10000) {
            *seg_out = default_cs;
        } else {
            *seg_out = (addr & 0xf0000) >> 4;
        }
        *off_out = (addr & 0xffff);
    }
    return 0;
}

static void cls(void) {
    printf("\033[2J\033[H");
    fflush(stdout);
}

static void dump_all_state(void) {
    cpu_dump_state();
    dma_dump_state();
    timer_dump_state();
    periph_dump_state();
    pic_dump_state();
    mda_dump_state();
    diskette_dump_state();
}

/* remove trailing newline and return it if it is there
   otherwise return 0
*/
static char chomp(char* s) {
    int end = strcspn(s, "\n");
    char c = s[end];
    s[end] = 0;
    return c == '\n';
}

static int unassemble_next_byte(byte* out, void* data) {
    dword* segoff = (dword*)data;
    word seg = HW(*segoff);
    word off = LW(*segoff);
    *out = load8(seg, off);
    off++;
    *segoff = pack32(seg, off);
    return 0;
}

static void debug_unassemble(char* arg) {
    int last_segoff;
    word seg = cpu_get_cs();
    word off = cpu_get_ip();
    word i = 0;
    struct a86d_ins ins;
    char outbuf[80];
    dword segoff;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (*arg) {
        if (segoff_parse(arg, &seg, &off, seg) < 0) {
            puts("Invalid address");
            return;
        }
    }
    segoff = pack32(seg, off);
    for (i = 0; i < 10; i++) {
        outbuf[0] = 0;
        last_segoff = segoff;
        a86d_decode(&ins, unassemble_next_byte, &segoff);
        a86d_rel_to_abs(&ins, LW(segoff));
        a86d_disasm(&ins, outbuf);
        printf("%04hx:%04hx: %s\n", HW(last_segoff), LW(last_segoff), outbuf);
    }
}

static int debug_go(char* arg) {
    word seg;
    word off;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (*arg) {
        if (segoff_parse(arg, &seg, &off, cpu_get_cs()) < 0) {
            puts("Invalid address");
            return 1;
        }
        cpu_set_cs(seg);
        cpu_set_ip(off);
    }
    cls();
    mda_refresh(1);
    term_set_raw();
    return 0;
}

static void debug_set_bp(char* arg) {
    word seg;
    word off;
    int bp;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (!*arg) {
        breakpoint_print();
        return;
    }
    if (segoff_parse(arg, &seg, &off, cpu_get_cs()) < 0) {
        puts("Invalid address");
        return;
    }
    bp = breakpoint_set(seg, off);
    if (bp < 0) {
        printf("No free breakpoints left\n");
    } else {
        printf("Breakpoint %i set at %04hx:%04hx\n", bp, seg, off);
    }
}

static void debug_clear_bp(char* arg) {
    dword bp;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (!*arg) {
        breakpoint_print();
        return;
    }
    if (strtou32(arg, &bp) < 0 || bp > BREAKPOINT_MAX) {
        puts("Invalid breakpoint number");
        return;
    }
    breakpoint_clear(bp);
    printf("Breakpoint %u cleared\n", bp);
}

static void debug_disk(char* arg) {
    byte drive;
    char* argend;
    /* skip leading space */
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (!*arg) {
        printf("Need drive number");
        return;
    }
    argend = arg;
    while (*argend && *argend != ' ') {
        argend++;
    }
    if (!*argend) {
        printf("Need path to disk\n");
        return;
    }
    *argend = 0;
    argend++;
    /* skip trailing space */
    while (*argend && *argend == ' ') {
        argend++;
    }
    if (!*argend) {
        printf("Need path to disk\n");
        return;
    }
    if (strtou8(arg, &drive) < 0 || drive > 4) {
        puts("Invalid drive number");
        return;
    }
    arg = argend;
    while (*argend && *argend != ' ') {
        argend++;
    }
    if (*argend) {
        *argend = 0;
    }
    if (diskette_insert(arg, drive, 0) < 0) {
        printf("Disk not inserted\n");
    }
}

static void debug_eject(char* arg) {
    byte drive;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (!*arg) {
        printf("Need drive number");
        return;
    }
    if (strtou8(arg, &drive) < 0 || drive > 4) {
        puts("Invalid drive number");
        return;
    }
    diskette_eject(drive);
}

static void debug_trace(char* arg) {
    dword count = 1;
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (*arg && strtou32(arg, &count) < 0) {
        puts("Invalid count");
        return;
    }
    while (count) {
        cpu_execute();
        cpu_fetch();
        count--;
    }
    dump_all_state();
}

static void debug_dump(char* arg) {
    word seg = cpu_get_cs();
    word off = cpu_get_ip();
    while (*arg && *arg == ' ') {
        arg++;
    }
    if (*arg) {
        if (segoff_parse(arg, &seg, &off, seg) < 0) {
            puts("Invalid address");
            return;
        }
    }
    printf("%04hx:%04hx\n", seg, off);
    memory_dump_at(seg, off);
}

int debug_cmd(void) {
    char line[BUFSIZ] = {0};
    char* lineptr = line;
    static char prev_line[BUFSIZ] = {0};
    char cmd;
    term_reset();
    printf("\n<> ");
    fflush(stdout);
    if (!fgets(line, BUFSIZ, stdin)) {
        puts("");
        if (ferror(stdin)) {
            perror(NULL);
            exit(1);
        }
        exit(0);
    }
    chomp(line);
    if (line[0] == '\0') {
        memcpy(line, prev_line, BUFSIZ);
    } else {
        memcpy(prev_line, line, BUFSIZ);
    }
    cmd = line[0] ? line[0] : 'r';
    lineptr += 1;
    switch (cmd) {
    case 'b':
        debug_set_bp(lineptr);
        break;
    case 'B':
        debug_clear_bp(lineptr);
        break;
    case 'c': /* clear */
        cls();
        break;
    case 'd': /* dump */
        debug_dump(lineptr);
        break;
    case 'D': /* disk */
        debug_disk(lineptr);
        break;
    case 'E': /* eject */
        debug_eject(lineptr);
        break;
    case 'g': /* go */
        return debug_go(lineptr);
    case 'q': /* quit */
        exit(0);
    case 'r': /* regs */
        dump_all_state();
        break;
    case 't': /* trace */
        debug_trace(lineptr);
        break;
    case 'u': /* unassemble */
        debug_unassemble(lineptr);
        break;
    case 'z': /* reboot */
        cpu_reset();
        cpu_fetch();
        printf("cpu reset\n");
        break;
    case '?':
        puts("?         help");
        puts("b #       break at address");
        puts("B #       clear breakpoint");
        puts("c         clear screen");
        puts("d [#]     dump [at address]");
        puts("D # path  insert disk in drive");
        puts("E #       eject disk from drive");
        puts("g [#]     go [to address]");
        puts("q         quit");
        puts("r         registers");
        puts("t [#]     trace [number of instructions]");
        puts("u [#]     unassemble [at address]");
        puts("z         reset CPU");
        break;
    default:
        puts("Unknown command");
        break;
    }
    return 1;
}
