/*
86d, an 8086 disassembler

Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <stdlib.h>
#include <string.h>

#include "86d.h"

/* used internally by the decoder only. */
enum decoder_optype {
    OT_N = 0, /* none */
    OT_1 = 1, /* constant one, used by shift/rotate ops. */
    OT_E = 2, /* memory or register (mod/rm) */
    OT_G = 3, /* general register (from middle 3 bits of modrm byte) */
    OT_I = 4, /* immediate */
    OT_O = 5, /* direct address, 16 bits. */
    OT_R = 6, /* relative jump/call */
    OT_S = 7, /* segreg */

    /* the values are significant as they are used in arithmetic */
    OT_AL = 8,
    OT_CL = 9,
    OT_DL = 10,
    OT_BL = 11,
    OT_AH = 12,
    OT_CH = 13,
    OT_DH = 14,
    OT_BH = 15,
    OT_AX = 16,
    OT_CX = 17,
    OT_DX = 18,
    OT_BX = 19,
    OT_SP = 20,
    OT_BP = 21,
    OT_SI = 22,
    OT_DI = 23,
    OT_ES = 24,
    OT_CS = 25,
    OT_SS = 26,
    OT_DS = 27
};

struct decoder_state {
    struct a86d_ins* ins;
    byte truncated; /* 1 if truncated. */
    void* data;
};

static int dummy(byte* unused, void* unused2) {
    (void)unused, (void)unused2;
    return -1;
}

static int (*next_byte)(byte*, void*) = dummy;

static word sign_extend16(byte in) {
    word tmp = in;
    if (in & 0x80) {
        tmp |= 0xff00;
    } else {
        tmp &=~ 0xff00;
    }
    return tmp;
}

/* Fetch immediate.
   takecount must be greater than 0 and no greater than 4.
 */
static dword immed(struct decoder_state* st, byte takecount) {
    struct a86d_ins* ins = st->ins;
    dword rval = 0;
    byte b;
    dword i;
    if (st->truncated) {
        return 0;
    }
    for (i = 0; takecount; i++, takecount--) {
        if (next_byte(&b, st->data) < 0) {
            /* not enough bytes to satisfy. */
            st->truncated = 1;
            return 0;
        }
        ins->length++;
        /* OR the byte into the result */
        rval |= ((dword)b << (i * 8));
    }
    return rval;
}

static byte immed8(struct decoder_state* st) {
    return immed(st, 1);
}

static word immed16(struct decoder_state* st) {
    return immed(st, 2);
}

static dword immed32(struct decoder_state* st) {
    return immed(st, 4);
}

static byte decode_mod(byte modrm) {
    return (modrm & 0xc0) >> 6; /* 11000000 */
}

static byte decode_rm(byte modrm) {
    return (modrm & 0x07);      /* 00000111 */
}

static byte decode_mid3(byte modrm) {
    return (modrm & 0x38) >> 3; /* 00111000 */
}

static void modrm_to_ins(struct decoder_state* st, byte modrm, byte whichop) {
    struct a86d_ins* ins = st->ins;
    byte mod, rm;
    word* disp = &(ins->disp);
    byte* dispsize = &(ins->dispsize);
    dword* operand = &(ins->operand[whichop]);
    enum a86d_oper_type* ot = &(ins->optype[whichop]);
    mod = decode_mod(modrm);
    rm = decode_rm(modrm);
    switch (mod & 0x3) {
    case 0: /* no disp */
        *disp = 0;
        *dispsize = 0;
        break;
    case 1: /* 8 bits sign-extended to 16 bits */
        *dispsize = 8;
        *disp = sign_extend16(immed8(st));
        break;
    case 2: /* 16 bits */
        *dispsize = 16;
        *disp = immed16(st);
        break;
    case 3: /* register */
        *disp = 0;
        *operand = rm;
        *ot = a86d_oper_reg;
        return;
    default:
        abort();
    }
    if (rm == 6 && !mod) {
        /* special case direct address */
        *operand = immed16(st);
        *ot = a86d_oper_directaddr;
    } else {
        *operand = rm;
        *ot = a86d_oper_mem;
    }
}

void makeoperand(struct decoder_state *st, enum decoder_optype t,
                 byte modrm, byte w, byte s) {
    struct a86d_ins* ins = st->ins;
    byte mid3 = 0;
    if (t == OT_E || t == OT_G || t == OT_S) {
        mid3 = decode_mid3(modrm);
    }
    ins->opsize[w] = s;
    switch (t) {
    case OT_N: ins->optype[w] = a86d_oper_none; break;
    case OT_E: modrm_to_ins(st, modrm, w); break;
    case OT_G: /* general register */
        ins->optype[w] = a86d_oper_reg;
        ins->operand[w] = mid3;
        break;
    /* specific regs */
    case OT_AX:
    case OT_CX:
    case OT_DX:
    case OT_BX:
    case OT_SP:
    case OT_BP:
    case OT_SI:
    case OT_DI:
        ins->optype[w] = a86d_oper_reg;
        ins->operand[w] = t - OT_AX;
        break;
    case OT_AL:
    case OT_CL:
    case OT_DL:
    case OT_BL:
    case OT_AH:
    case OT_CH:
    case OT_DH:
    case OT_BH:
        ins->optype[w] = a86d_oper_reg;
        ins->operand[w] = t - OT_AL;
        break;
    case OT_ES:
    case OT_CS:
    case OT_SS:
    case OT_DS:
        ins->optype[w] = a86d_oper_segreg;
        ins->operand[w] = t - OT_ES;
        break;
    case OT_I: /* immed */
    case OT_R:
        if (t == OT_R) { /* rel imm */
            ins->optype[w] = a86d_oper_relimm;
        } else { /* imm */
            ins->optype[w] = a86d_oper_imm;
        }
        if (s == 8) {
            ins->operand[w] = immed8(st);
        } else if (s == 16) {
            ins->operand[w] = immed16(st);
        } else if (s == 32) {
            ins->operand[w] = immed32(st);
        } else {
            /* invalid operand size. */
            /* only a bug in this file would cause this */
            abort();
        }
        break;
    case OT_S: /* segreg */
        ins->optype[w] = a86d_oper_segreg;
        ins->operand[w] = mid3;
        break;
    case OT_O: /* direct address, 16 bits */
        ins->optype[w] = a86d_oper_directaddr;
        ins->operand[w] = immed16(st);
        break;
    case OT_1: /* constant one (used for shifts/rotates) */
        ins->optype[w] = a86d_oper_imm;
        ins->operand[w] = 1;
        break;
    default:
        abort();
    }
}

static void make2_with_modrm(struct decoder_state* st, enum a86d_op op,
                    enum decoder_optype t1, enum decoder_optype t2,
                    byte s1, byte s2, byte modrm) {
    st->ins->op = op;
    makeoperand(st, t1, modrm, 0, s1);
    makeoperand(st, t2, modrm, 1, s2);
}

static void make2(struct decoder_state* st, enum a86d_op op,
                    enum decoder_optype t1, enum decoder_optype t2,
                    byte s1, byte s2) {
    byte modrm = 0;
    if (t1 == OT_E || t1 == OT_G || t1 == OT_S
     || t2 == OT_E || t2 == OT_G || t2 == OT_S) {
        modrm = immed8(st);
    }
    make2_with_modrm(st, op, t1, t2, s1, s2, modrm);
}

static void make_with_modrm(struct decoder_state* st, enum a86d_op op,
             enum decoder_optype t1, enum decoder_optype t2,
             byte s, byte modrm) {
    make2_with_modrm(st, op, t1, t2, s, s, modrm);
}

static void make(struct decoder_state* st, enum a86d_op op,
             enum decoder_optype t1, enum decoder_optype t2,
             byte s) {
    make2(st, op, t1, t2, s, s);
}

static void makegrp1(struct decoder_state* st, byte s1, byte s2) {
    byte modrm = immed8(st);
    byte mid3 = decode_mid3(modrm);
    switch (mid3 & 0x7) {
    case 0: make2_with_modrm(st, a86d_op_add, OT_E, OT_I, s1, s2, modrm); break;
    case 1: make2_with_modrm(st, a86d_op_or, OT_E, OT_I, s1, s2, modrm); break;
    case 2: make2_with_modrm(st, a86d_op_adc, OT_E, OT_I, s1, s2, modrm); break;
    case 3: make2_with_modrm(st, a86d_op_sbb, OT_E, OT_I, s1, s2, modrm); break;
    case 4: make2_with_modrm(st, a86d_op_and, OT_E, OT_I, s1, s2, modrm); break;
    case 5: make2_with_modrm(st, a86d_op_sub, OT_E, OT_I, s1, s2, modrm); break;
    case 6: make2_with_modrm(st, a86d_op_xor, OT_E, OT_I, s1, s2, modrm); break;
    case 7: make2_with_modrm(st, a86d_op_cmp, OT_E, OT_I, s1, s2, modrm); break;
    default: abort();
    }
    if (s1 == 16 && s2 == 8) {
        /* sign-extend 8-bit immediate to 16 bits. */
        st->ins->opsize[1] = 16;
        st->ins->operand[1] = (dword)sign_extend16(st->ins->operand[1]);
    }
}

static void makegrp2(struct decoder_state* st,
              enum decoder_optype arg2,
              byte s1) {
    byte modrm = immed8(st);
    byte mid3 = decode_mid3(modrm);
    switch (mid3 & 0x7) {
    case 0: make2_with_modrm(st, a86d_op_rol, OT_E, arg2, s1, 8, modrm); break;
    case 1: make2_with_modrm(st, a86d_op_ror, OT_E, arg2, s1, 8, modrm); break;
    case 2: make2_with_modrm(st, a86d_op_rcl, OT_E, arg2, s1, 8, modrm); break;
    case 3: make2_with_modrm(st, a86d_op_rcr, OT_E, arg2, s1, 8, modrm); break;
    case 4: make2_with_modrm(st, a86d_op_shl, OT_E, arg2, s1, 8, modrm); break;
    case 5: make2_with_modrm(st, a86d_op_shr, OT_E, arg2, s1, 8, modrm); break;
    /* case 6 is an undocumented combination */
    case 6: make2_with_modrm(st, a86d_op_db, OT_E, arg2, s1, 8, modrm); break;
    case 7: make2_with_modrm(st, a86d_op_sar, OT_E, arg2, s1, 8, modrm); break;
    default: abort();
    }
}

void makegrp3(struct decoder_state* st, byte s) {
    byte modrm = immed8(st);
    byte mid3 = decode_mid3(modrm);
    switch (mid3 & 0x7) {
    case 0: make_with_modrm(st, a86d_op_test, OT_E, OT_I, s, modrm); break;
    /* case 1 is an undocumented combination */
    case 1: make_with_modrm(st, a86d_op_db, OT_E, OT_N, s, modrm); break;
    case 2: make_with_modrm(st, a86d_op_not, OT_E, OT_N, s, modrm); break;
    case 3: make_with_modrm(st, a86d_op_neg, OT_E, OT_N, s, modrm); break;
    case 4: make_with_modrm(st, a86d_op_mul, OT_E, OT_N, s, modrm); break;
    case 5: make_with_modrm(st, a86d_op_imul, OT_E, OT_N, s, modrm); break;
    case 6: make_with_modrm(st, a86d_op_div, OT_E, OT_N, s, modrm); break;
    case 7: make_with_modrm(st, a86d_op_idiv, OT_E, OT_N, s, modrm); break;
    default: abort();
    }
}

static void makegrp4(struct decoder_state* st) {
    byte modrm = immed8(st);
    byte mid3 = decode_mid3(modrm);
    switch (mid3 & 0x7) {
    case 0: make_with_modrm(st, a86d_op_inc, OT_E, OT_N, 8, modrm); break;
    case 1: make_with_modrm(st, a86d_op_dec, OT_E, OT_N, 8, modrm); break;
    /* case 2 through 7 are undocumented combinations */
    case 2: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    case 3: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    case 4: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    case 5: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    case 6: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    case 7: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 8, modrm); break;
    default: abort();
    }
}

static void makegrp5(struct decoder_state* st) {
    byte modrm = immed8(st);
    byte mid3 = decode_mid3(modrm);
    switch (mid3 & 0x7) {
    case 0: make_with_modrm(st, a86d_op_inc, OT_E, OT_N, 16, modrm); break;
    case 1: make_with_modrm(st, a86d_op_dec, OT_E, OT_N, 16, modrm); break;
    case 2: make_with_modrm(st, a86d_op_call_mem16, OT_E, OT_N, 16, modrm); break;
    case 3: make_with_modrm(st, a86d_op_call_mem32, OT_E, OT_N, 32, modrm); break;
    case 4: make_with_modrm(st, a86d_op_jmp_mem16, OT_E, OT_N, 16, modrm); break;
    case 5: make_with_modrm(st, a86d_op_jmp_mem32, OT_E, OT_N, 32, modrm); break;
    case 6: make_with_modrm(st, a86d_op_push, OT_E, OT_N, 16, modrm); break;
    /* case 7 is an undocumented combination. */
    case 7: make_with_modrm(st, a86d_op_db, OT_E, OT_N, 16, modrm); break;
    default: abort();
    }
}

size_t a86d_decode(struct a86d_ins* ins,
                   int (*nextbyte)(byte*, void*), void* data) {
    struct decoder_state state = {0};
    struct decoder_state* st = &state;
    byte is_prefix;
    byte opcode;

    if (nextbyte) {
        next_byte = nextbyte;
    }

    /* init */
    st->ins = ins;
    st->truncated = 0;
    st->data = data;
    ins->lock = 0;
    ins->rep_mode = a86d_no_rep;
    ins->op = a86d_op_none;
    ins->seg = a86d_seg_none;
    ins->length = 0;
    ins->optype[0] = a86d_oper_none;
    ins->optype[1] = a86d_oper_none;
    ins->operand[0] = 0;
    ins->operand[1] = 0;
    ins->opsize[0] = 0;
    ins->opsize[1] = 0;
    ins->disp = 0;
    ins->dispsize = 0;

next_op:
    is_prefix = 1;
    opcode = immed8(st);
    /* prefixes */
    switch (opcode) {
    case 0x26: ins->seg = a86d_seg_es; break;
    case 0x2e: ins->seg = a86d_seg_cs; break;
    case 0x36: ins->seg = a86d_seg_ss; break;
    case 0x3e: ins->seg = a86d_seg_ds; break;
    case 0xf0: ins->lock = 1; break;
    case 0xf1: ins->lock = 1; break;
    case 0xf2: ins->rep_mode = a86d_repne; break;
    case 0xf3: ins->rep_mode = a86d_rep; break;
    default: is_prefix = 0; break;
    }
    if (is_prefix) {
        goto next_op;
    }
    switch (opcode) {
    case 0x00: make(st, a86d_op_add, OT_E, OT_G, 8);  break;  /* add Eb, Gb */
    case 0x01: make(st, a86d_op_add, OT_E, OT_G, 16); break;  /* add Ew, Gw */
    case 0x02: make(st, a86d_op_add, OT_G, OT_E, 8);  break;  /* add Gb, Eb */
    case 0x03: make(st, a86d_op_add, OT_G, OT_E, 16); break;  /* add Gw, Ew */
    case 0x04: make(st, a86d_op_add, OT_AL, OT_I, 8);  break;  /* add al, Ib */
    case 0x05: make(st, a86d_op_add, OT_AX, OT_I, 16); break;  /* add ax, Iw */
    case 0x06: make(st, a86d_op_push, OT_ES, OT_N, 16); break; /* push es */
    case 0x07: make(st, a86d_op_pop, OT_ES, OT_N, 16);  break;  /* pop es */
    case 0x08: make(st, a86d_op_or, OT_E, OT_G, 8);  break;  /* or Eb, Gb */
    case 0x09: make(st, a86d_op_or, OT_E, OT_G, 16); break;  /* or Ew, Gw */
    case 0x0a: make(st, a86d_op_or, OT_G, OT_E, 8);  break;  /* or Gb, Eb */
    case 0x0b: make(st, a86d_op_or, OT_G, OT_E, 16); break;  /* or Gw, Ew */
    case 0x0c: make(st, a86d_op_or, OT_AL, OT_I, 8);  break;  /* or al, Ib */
    case 0x0d: make(st, a86d_op_or, OT_AX, OT_I, 16); break;  /* or ax, Iw */
    case 0x0e: make(st, a86d_op_push, OT_CS, OT_N, 16); break; /* push cs */
    /* pop cs only valid on 8086/8088. */
    case 0x0f: make(st, a86d_op_pop, OT_CS, OT_N, 16);  break;  /* pop cs */
    case 0x10: make(st, a86d_op_adc, OT_E, OT_G, 8);  break;  /* add Eb, Gb */
    case 0x11: make(st, a86d_op_adc, OT_E, OT_G, 16); break;  /* add Ew, Gw */
    case 0x12: make(st, a86d_op_adc, OT_G, OT_E, 8);  break;  /* add Gb, Eb */
    case 0x13: make(st, a86d_op_adc, OT_G, OT_E, 16); break;  /* add Gw, Ew */
    case 0x14: make(st, a86d_op_adc, OT_AL, OT_I, 8);  break;  /* add al, Ib */
    case 0x15: make(st, a86d_op_adc, OT_AX, OT_I, 16); break;  /* add ax, Iw */
    case 0x16: make(st, a86d_op_push, OT_SS, OT_N, 16); break; /* push ss */
    case 0x17: make(st, a86d_op_pop, OT_SS, OT_N, 16);  break;  /* pop ss */
    case 0x18: make(st, a86d_op_sbb, OT_E, OT_G, 8);  break;  /* sbb Eb, Gb */
    case 0x19: make(st, a86d_op_sbb, OT_E, OT_G, 16); break;  /* sbb Ew, Gw */
    case 0x1a: make(st, a86d_op_sbb, OT_G, OT_E, 8);  break;  /* sbb Gb, Eb */
    case 0x1b: make(st, a86d_op_sbb, OT_G, OT_E, 16); break;  /* sbb Gw, Ew */
    case 0x1c: make(st, a86d_op_sbb, OT_AL, OT_I, 8);  break;  /* sbb al, Ib */
    case 0x1d: make(st, a86d_op_sbb, OT_AX, OT_I, 16); break;  /* sbb ax, Iw */
    case 0x1e: make(st, a86d_op_push, OT_DS, OT_N, 16); break; /* push ds */
    case 0x1f: make(st, a86d_op_pop, OT_DS, OT_N, 16);  break;  /* pop ds */
    case 0x20: make(st, a86d_op_and, OT_E, OT_G, 8);  break;  /* and Eb, Gb */
    case 0x21: make(st, a86d_op_and, OT_E, OT_G, 16); break;  /* and Ew, Gw */
    case 0x22: make(st, a86d_op_and, OT_G, OT_E, 8);  break;  /* and Gb, Eb */
    case 0x23: make(st, a86d_op_and, OT_G, OT_E, 16); break;  /* and Gw, Ew */
    case 0x24: make(st, a86d_op_and, OT_AL, OT_I, 8);  break;  /* and al, Ib */
    case 0x25: make(st, a86d_op_and, OT_AX, OT_I, 16); break;  /* and ax, Iw */
    /* 0x26 es: */
    case 0x27: make(st, a86d_op_daa, OT_N, OT_N, 16); break;  /* daa */
    case 0x28: make(st, a86d_op_sub, OT_E, OT_G, 8);  break;  /* sub Eb, Gb */
    case 0x29: make(st, a86d_op_sub, OT_E, OT_G, 16); break;  /* sub Ew, Gw */
    case 0x2a: make(st, a86d_op_sub, OT_G, OT_E, 8);  break;  /* sub Gb, Eb */
    case 0x2b: make(st, a86d_op_sub, OT_G, OT_E, 16); break;  /* sub Gw, Ew */
    case 0x2c: make(st, a86d_op_sub, OT_AL, OT_I, 8);  break;  /* sub al, Ib */
    case 0x2d: make(st, a86d_op_sub, OT_AX, OT_I, 16); break;  /* sub ax, Iw */
    /* 0x2e cs: */
    case 0x2f: make(st, a86d_op_das, OT_N, OT_N, 16); break;  /* das */
    case 0x30: make(st, a86d_op_xor, OT_E, OT_G, 8);  break;  /* xor Eb, Gb */
    case 0x31: make(st, a86d_op_xor, OT_E, OT_G, 16); break;  /* xor Ew, Gw */
    case 0x32: make(st, a86d_op_xor, OT_G, OT_E, 8);  break;  /* xor Gb, Eb */
    case 0x33: make(st, a86d_op_xor, OT_G, OT_E, 16); break;  /* xor Gw, Ew */
    case 0x34: make(st, a86d_op_xor, OT_AL, OT_I, 8);  break;  /* xor al, Ib */
    case 0x35: make(st, a86d_op_xor, OT_AX, OT_I, 16); break;  /* xor ax, Iw */
    /* 0x36 ss: */
    case 0x37: make(st, a86d_op_aaa, OT_N, OT_N, 16); break;  /* aaa */
    case 0x38: make(st, a86d_op_cmp, OT_E, OT_G, 8);  break;  /* cmp Eb, Gb */
    case 0x39: make(st, a86d_op_cmp, OT_E, OT_G, 16); break;  /* cmp Ew, Gw */
    case 0x3a: make(st, a86d_op_cmp, OT_G, OT_E, 8);  break;  /* cmp Gb, Eb */
    case 0x3b: make(st, a86d_op_cmp, OT_G, OT_E, 16); break;  /* cmp Gw, Ew */
    case 0x3c: make(st, a86d_op_cmp, OT_AL, OT_I, 8);  break;  /* cmp al, Ib */
    case 0x3d: make(st, a86d_op_cmp, OT_AX, OT_I, 16); break;  /* cmp ax, Iw */
    /* 0x3e ds: */
    case 0x3f: make(st, a86d_op_aas, OT_N, OT_N, 16); break;  /* aas */
    case 0x40: make(st, a86d_op_inc, OT_AX, OT_N, 16); break; /* inc ax */
    case 0x41: make(st, a86d_op_inc, OT_CX, OT_N, 16); break; /* inc cx */
    case 0x42: make(st, a86d_op_inc, OT_DX, OT_N, 16); break; /* inc dx */
    case 0x43: make(st, a86d_op_inc, OT_BX, OT_N, 16); break; /* inc bx */
    case 0x44: make(st, a86d_op_inc, OT_SP, OT_N, 16); break; /* inc sp */
    case 0x45: make(st, a86d_op_inc, OT_BP, OT_N, 16); break; /* inc bp */
    case 0x46: make(st, a86d_op_inc, OT_SI, OT_N, 16); break; /* inc si */
    case 0x47: make(st, a86d_op_inc, OT_DI, OT_N, 16); break; /* inc di */
    case 0x48: make(st, a86d_op_dec, OT_AX, OT_N, 16); break; /* dec ax */
    case 0x49: make(st, a86d_op_dec, OT_CX, OT_N, 16); break; /* dec cx */
    case 0x4a: make(st, a86d_op_dec, OT_DX, OT_N, 16); break; /* dec dx */
    case 0x4b: make(st, a86d_op_dec, OT_BX, OT_N, 16); break; /* dec bx */
    case 0x4c: make(st, a86d_op_dec, OT_SP, OT_N, 16); break; /* dec sp */
    case 0x4d: make(st, a86d_op_dec, OT_BP, OT_N, 16); break; /* dec bp */
    case 0x4e: make(st, a86d_op_dec, OT_SI, OT_N, 16); break; /* dec si */
    case 0x4f: make(st, a86d_op_dec, OT_DI, OT_N, 16); break; /* dec di */
    case 0x50: make(st, a86d_op_push, OT_AX, OT_N, 16); break; /* push ax */
    case 0x51: make(st, a86d_op_push, OT_CX, OT_N, 16); break; /* push cx */
    case 0x52: make(st, a86d_op_push, OT_DX, OT_N, 16); break; /* push dx */
    case 0x53: make(st, a86d_op_push, OT_BX, OT_N, 16); break; /* push bx */
    case 0x54: make(st, a86d_op_push, OT_SP, OT_N, 16); break; /* push sp */
    case 0x55: make(st, a86d_op_push, OT_BP, OT_N, 16); break; /* push bp */
    case 0x56: make(st, a86d_op_push, OT_SI, OT_N, 16); break; /* push si */
    case 0x57: make(st, a86d_op_push, OT_DI, OT_N, 16); break; /* push di */
    case 0x58: make(st, a86d_op_pop, OT_AX, OT_N, 16); break; /* pop ax */
    case 0x59: make(st, a86d_op_pop, OT_CX, OT_N, 16); break; /* pop cx */
    case 0x5a: make(st, a86d_op_pop, OT_DX, OT_N, 16); break; /* pop dx */
    case 0x5b: make(st, a86d_op_pop, OT_BX, OT_N, 16); break; /* pop bx */
    case 0x5c: make(st, a86d_op_pop, OT_SP, OT_N, 16); break; /* pop sp */
    case 0x5d: make(st, a86d_op_pop, OT_BP, OT_N, 16); break; /* pop bp */
    case 0x5e: make(st, a86d_op_pop, OT_SI, OT_N, 16); break; /* pop si */
    case 0x5f: make(st, a86d_op_pop, OT_DI, OT_N, 16); break; /* pop di */
    /* Undocumented 8086 quirk, 0x6x byte decodes the same as 0x7x. */
    /* This is not true on later processors. */
    case 0x60: make(st, a86d_op_jo, OT_R, OT_N, 8); break;
    case 0x61: make(st, a86d_op_jno, OT_R, OT_N, 8); break;
    case 0x62: make(st, a86d_op_jc, OT_R, OT_N, 8); break;
    case 0x63: make(st, a86d_op_jnc, OT_R, OT_N, 8); break;
    case 0x64: make(st, a86d_op_jz, OT_R, OT_N, 8); break;
    case 0x65: make(st, a86d_op_jnz, OT_R, OT_N, 8); break;
    case 0x66: make(st, a86d_op_jbe, OT_R, OT_N, 8); break;
    case 0x67: make(st, a86d_op_ja, OT_R, OT_N, 8); break;
    case 0x68: make(st, a86d_op_js, OT_R, OT_N, 8); break;
    case 0x69: make(st, a86d_op_jns, OT_R, OT_N, 8); break;
    case 0x6a: make(st, a86d_op_jp, OT_R, OT_N, 8); break;
    case 0x6b: make(st, a86d_op_jpo, OT_R, OT_N, 8); break;
    case 0x6c: make(st, a86d_op_jl, OT_R, OT_N, 8); break;
    case 0x6d: make(st, a86d_op_jge, OT_R, OT_N, 8); break;
    case 0x6e: make(st, a86d_op_jle, OT_R, OT_N, 8); break;
    case 0x6f: make(st, a86d_op_jg, OT_R, OT_N, 8); break;
    case 0x70: make(st, a86d_op_jo, OT_R, OT_N, 8); break;
    case 0x71: make(st, a86d_op_jno, OT_R, OT_N, 8); break;
    case 0x72: make(st, a86d_op_jc, OT_R, OT_N, 8); break;
    case 0x73: make(st, a86d_op_jnc, OT_R, OT_N, 8); break;
    case 0x74: make(st, a86d_op_jz, OT_R, OT_N, 8); break;
    case 0x75: make(st, a86d_op_jnz, OT_R, OT_N, 8); break;
    case 0x76: make(st, a86d_op_jbe, OT_R, OT_N, 8); break;
    case 0x77: make(st, a86d_op_ja, OT_R, OT_N, 8); break;
    case 0x78: make(st, a86d_op_js, OT_R, OT_N, 8); break;
    case 0x79: make(st, a86d_op_jns, OT_R, OT_N, 8); break;
    case 0x7a: make(st, a86d_op_jp, OT_R, OT_N, 8); break;
    case 0x7b: make(st, a86d_op_jpo, OT_R, OT_N, 8); break;
    case 0x7c: make(st, a86d_op_jl, OT_R, OT_N, 8); break;
    case 0x7d: make(st, a86d_op_jge, OT_R, OT_N, 8); break;
    case 0x7e: make(st, a86d_op_jle, OT_R, OT_N, 8); break;
    case 0x7f: make(st, a86d_op_jg, OT_R, OT_N, 8); break;
    case 0x80: makegrp1(st, 8, 8); break;
    case 0x81: makegrp1(st, 16, 16); break;
    case 0x82: makegrp1(st, 8, 8); break;
    case 0x83: makegrp1(st, 16, 8); break;
    case 0x84: make(st, a86d_op_test, OT_E, OT_G, 8);  break;  /* test Eb, Gb */
    case 0x85: make(st, a86d_op_test, OT_E, OT_G, 16); break;  /* test Ew, Gw */
    case 0x86: make(st, a86d_op_xchg, OT_E, OT_G, 8);  break;  /* xchg Eb, Gb */
    case 0x87: make(st, a86d_op_xchg, OT_E, OT_G, 16); break;  /* xchg Ew, Gw */
    case 0x88: make(st, a86d_op_mov, OT_E, OT_G, 8);  break;  /* mov Eb, Gb */
    case 0x89: make(st, a86d_op_mov, OT_E, OT_G, 16); break;  /* mov Ew, Gw */
    case 0x8a: make(st, a86d_op_mov, OT_G, OT_E, 8);  break;  /* mov Gb, Eb */
    case 0x8b: make(st, a86d_op_mov, OT_G, OT_E, 16); break;  /* mov Gw, Ew */
    case 0x8c: make(st, a86d_op_mov, OT_E, OT_S, 16); break;  /* mov Ew, Sw */
    case 0x8d: make(st, a86d_op_lea, OT_G, OT_E, 16); break;  /* lea */
    case 0x8e: make(st, a86d_op_mov, OT_S, OT_E, 16); break;  /* mov Sw, Ew */
    case 0x8f: make(st, a86d_op_pop, OT_E, OT_N, 16); break;  /* pop Ew */
    case 0x90: make(st, a86d_op_nop, OT_N, OT_N, 16); break;
    case 0x91: make(st, a86d_op_xchg, OT_AX, OT_CX, 16); break;
    case 0x92: make(st, a86d_op_xchg, OT_AX, OT_DX, 16); break;
    case 0x93: make(st, a86d_op_xchg, OT_AX, OT_BX, 16); break;
    case 0x94: make(st, a86d_op_xchg, OT_AX, OT_SP, 16); break;
    case 0x95: make(st, a86d_op_xchg, OT_AX, OT_BP, 16); break;
    case 0x96: make(st, a86d_op_xchg, OT_AX, OT_SI, 16); break;
    case 0x97: make(st, a86d_op_xchg, OT_AX, OT_DI, 16); break;
    case 0x98: make(st, a86d_op_cbw, OT_N, OT_N, 0); break;
    case 0x99: make(st, a86d_op_cwd, OT_N, OT_N, 0); break;
    case 0x9a: make(st, a86d_op_call_abs32, OT_I, OT_N, 32); break;
    case 0x9b: make(st, a86d_op_wait, OT_N, OT_N, 0); break;
    case 0x9c: make(st, a86d_op_pushf, OT_N, OT_N, 0); break;
    case 0x9d: make(st, a86d_op_popf, OT_N, OT_N, 0); break;
    case 0x9e: make(st, a86d_op_sahf, OT_N, OT_N, 0); break;
    case 0x9f: make(st, a86d_op_lahf, OT_N, OT_N, 0); break;
    case 0xa0: make(st, a86d_op_mov, OT_AL, OT_O, 8); break;
    case 0xa1: make(st, a86d_op_mov, OT_AX, OT_O, 16); break;
    case 0xa2: make(st, a86d_op_mov, OT_O, OT_AL, 8); break;
    case 0xa3: make(st, a86d_op_mov, OT_O, OT_AX, 16); break;
    case 0xa4: make(st, a86d_op_movsb, OT_N, OT_N, 0); break;
    case 0xa5: make(st, a86d_op_movsw, OT_N, OT_N, 0); break;
    case 0xa6: make(st, a86d_op_cmpsb, OT_N, OT_N, 0); break;
    case 0xa7: make(st, a86d_op_cmpsw, OT_N, OT_N, 0); break;
    case 0xa8: make(st, a86d_op_test, OT_AL, OT_I, 8); break;
    case 0xa9: make(st, a86d_op_test, OT_AX, OT_I, 16); break;
    case 0xaa: make(st, a86d_op_stosb, OT_N, OT_N, 0); break;
    case 0xab: make(st, a86d_op_stosw, OT_N, OT_N, 0); break;
    case 0xac: make(st, a86d_op_lodsb, OT_N, OT_N, 0); break;
    case 0xad: make(st, a86d_op_lodsw, OT_N, OT_N, 0); break;
    case 0xae: make(st, a86d_op_scasb, OT_N, OT_N, 0); break;
    case 0xaf: make(st, a86d_op_scasw, OT_N, OT_N, 0); break;
    case 0xb0: make(st, a86d_op_mov, OT_AL, OT_I, 8); break;
    case 0xb1: make(st, a86d_op_mov, OT_CL, OT_I, 8); break;
    case 0xb2: make(st, a86d_op_mov, OT_DL, OT_I, 8); break;
    case 0xb3: make(st, a86d_op_mov, OT_BL, OT_I, 8); break;
    case 0xb4: make(st, a86d_op_mov, OT_AH, OT_I, 8); break;
    case 0xb5: make(st, a86d_op_mov, OT_CH, OT_I, 8); break;
    case 0xb6: make(st, a86d_op_mov, OT_DH, OT_I, 8); break;
    case 0xb7: make(st, a86d_op_mov, OT_BH, OT_I, 8); break;
    case 0xb8: make(st, a86d_op_mov, OT_AX, OT_I, 16); break;
    case 0xb9: make(st, a86d_op_mov, OT_CX, OT_I, 16); break;
    case 0xba: make(st, a86d_op_mov, OT_DX, OT_I, 16); break;
    case 0xbb: make(st, a86d_op_mov, OT_BX, OT_I, 16); break;
    case 0xbc: make(st, a86d_op_mov, OT_SP, OT_I, 16); break;
    case 0xbd: make(st, a86d_op_mov, OT_BP, OT_I, 16); break;
    case 0xbe: make(st, a86d_op_mov, OT_SI, OT_I, 16); break;
    case 0xbf: make(st, a86d_op_mov, OT_DI, OT_I, 16); break;
    case 0xc0: make(st, a86d_op_ret, OT_I, OT_N, 16); break; /* 8086 only */
    case 0xc1: make(st, a86d_op_ret, OT_N, OT_N, 0); break; /* 8086 only */
    case 0xc2: make(st, a86d_op_ret, OT_I, OT_N, 16); break;
    case 0xc3: make(st, a86d_op_ret, OT_N, OT_N, 0); break;
    case 0xc4: make2(st, a86d_op_les, OT_G, OT_E, 16, 32); break;
    case 0xc5: make2(st, a86d_op_lds, OT_G, OT_E, 16, 32); break;
    case 0xc6: make(st, a86d_op_mov, OT_E, OT_I, 8); break;
    case 0xc7: make(st, a86d_op_mov, OT_E, OT_I, 16); break;
    case 0xc8: make(st, a86d_op_retf, OT_I, OT_N, 16); break; /* 8086 only */
    case 0xc9: make(st, a86d_op_retf, OT_N, OT_N, 0); break;  /* 8086 only */
    case 0xca: make(st, a86d_op_retf, OT_I, OT_N, 16); break;
    case 0xcb: make(st, a86d_op_retf, OT_N, OT_N, 0); break;
    case 0xcc: make(st, a86d_op_int3, OT_N, OT_N, 0); break;
    case 0xcd: make(st, a86d_op_int, OT_I, OT_N, 8); break;
    case 0xce: make(st, a86d_op_into, OT_N, OT_N, 0); break;
    case 0xcf: make(st, a86d_op_iret, OT_N, OT_N, 0); break;
    case 0xd0: makegrp2(st, OT_1, 8); break;
    case 0xd1: makegrp2(st, OT_1, 16); break;
    case 0xd2: makegrp2(st, OT_CL, 8); break;
    case 0xd3: makegrp2(st, OT_CL, 16); break;
    case 0xd4: make(st, a86d_op_aam, OT_I, OT_N, 8); break; /* undocumented arg */
    case 0xd5: make(st, a86d_op_aad, OT_I, OT_N, 8); break; /* undocumented arg */
    case 0xd6: make(st, a86d_op_salc, OT_N, OT_N, 0); break; /* undocumented opcode */
    case 0xd7: make(st, a86d_op_xlat, OT_N, OT_N, 0); break;
    case 0xD8:
    case 0xD9:
    case 0xDA:
    case 0xDB:
    case 0xDC:
    case 0xDD:
    case 0xDE:
    case 0xDF:
        make(st, a86d_op_esc, OT_E, OT_N, 16); break;
    case 0xe0: make(st, a86d_op_loopnz, OT_R, OT_N, 8); break;
    case 0xe1: make(st, a86d_op_loopz, OT_R, OT_N, 8); break;
    case 0xe2: make(st, a86d_op_loop, OT_R, OT_N, 8); break;
    case 0xe3: make(st, a86d_op_jcxz, OT_R, OT_N, 8); break;
    case 0xe4: make(st, a86d_op_in, OT_AL, OT_I, 8); break;
    case 0xe5: make2(st, a86d_op_in, OT_AX, OT_I, 16, 8); break;
    case 0xe6: make(st, a86d_op_out, OT_I, OT_AL, 8); break;
    case 0xe7: make2(st, a86d_op_out, OT_I, OT_AX, 8, 16); break;
    case 0xe8: make(st, a86d_op_call_rel16, OT_R, OT_N, 16); break;
    case 0xe9: make(st, a86d_op_jmp_rel16, OT_R, OT_N, 16); break;
    case 0xea: make(st, a86d_op_jmp_abs32, OT_I, OT_N, 32); break;
    case 0xeb: make(st, a86d_op_jmp_rel8, OT_R, OT_N, 8); break;
    case 0xec: make2(st, a86d_op_in, OT_AL, OT_DX, 8, 16); break;
    case 0xed: make(st, a86d_op_in, OT_AX, OT_DX, 16); break;
    case 0xee: make2(st, a86d_op_out, OT_DX, OT_AL, 16, 8); break;
    case 0xef: make(st, a86d_op_out, OT_DX, OT_AX, 16); break;
    /* 0xf0 lock */
    /* 0xf1 lock (int1/icebp on later processors) */
    /* 0xf2 rep / repne */
    /* 0xf3 rep / repe */
    case 0xf4: make(st, a86d_op_hlt, OT_N, OT_N, 0); break;
    case 0xf5: make(st, a86d_op_cmc, OT_N, OT_N, 0); break;
    case 0xf6: makegrp3(st, 8); break;
    case 0xf7: makegrp3(st, 16); break;
    case 0xf8: make(st, a86d_op_clc, OT_N, OT_N, 0); break;
    case 0xf9: make(st, a86d_op_stc, OT_N, OT_N, 0); break;
    case 0xfa: make(st, a86d_op_cli, OT_N, OT_N, 0); break;
    case 0xfb: make(st, a86d_op_sti, OT_N, OT_N, 0); break;
    case 0xfc: make(st, a86d_op_cld, OT_N, OT_N, 0); break;
    case 0xfd: make(st, a86d_op_std, OT_N, OT_N, 0); break;
    case 0xfe: makegrp4(st); break;
    case 0xff: makegrp5(st); break;
    default: abort();
    }
    if (st->truncated) {
        /* fall back to db */
        ins->op = a86d_op_db;
    }
    return ins->length;
}

void a86d_rel_to_abs(struct a86d_ins* ins, word offset) {
    int i;
    for (i = 0; i < 2; i++) {
        if (ins->optype[i] == a86d_oper_relimm) {
            if (ins->opsize[i] == 8) {
                ins->operand[i] = sign_extend16(ins->operand[i]);
                ins->opsize[i] = 16;
            }
            ins->operand[i] += offset;
        }
    }
}
