/*
86d, an 8086 disassembler

Written by T. Taylor Gurney, 2022-2024.
Public domain.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "86d.h"

#define min(a, b) ((a) < (b) ? (a) : (b))

/* convert enum ins_code to mnemonic by index. */
static char* mnemonic_table[] = {
    "(none)",
    "aaa",
    "aad",
    "aam",
    "aas",
    "adc",
    "add",
    "and",
    "call",
    "call near",
    "call far",
    "call",
    "cbw",
    "clc",
    "cld",
    "cli",
    "cmc",
    "cmp",
    "cmpsb",
    "cmpsw",
    "cwd",
    "daa",
    "das",
    "dec",
    "div",
    "esc",
    "hlt",
    "idiv",
    "imul",
    "in",
    "inc",
    "int",
    "int3",
    "into",
    "iret",
    "ja",
    "jbe",
    "jc",
    "jcxz",
    "jz",
    "jg",
    "jge",
    "jl",
    "jle",
    "jmp",
    "jmp near",
    "jmp far",
    "jmp",
    "jmp short",
    "jnc",
    "jnz",
    "jno",
    "jns",
    "jo",
    "jp",
    "jpo",
    "js",
    "lahf",
    "lds",
    "lea",
    "les",
    "lodsb",
    "lodsw",
    "loop",
    "loopnz",
    "loopz",
    "mov",
    "movsb",
    "movsw",
    "mul",
    "neg",
    "nop",
    "not",
    "or",
    "out",
    "pop",
    "popf",
    "push",
    "pushf",
    "rcl",
    "rcr",
    "ret",
    "retf",
    "rol",
    "ror",
    "sahf",
    "salc",
    "sar",
    "sbb",
    "scasb",
    "scasw",
    "shl",
    "shr",
    "stc",
    "std",
    "sti",
    "stosb",
    "stosw",
    "sub",
    "test",
    "wait",
    "xchg",
    "xlat",
    "xor",
    "db"
};

static char* decode_reg16(byte bits3) {
    switch (bits3 & 0x7) {
    case 0: return "ax";
    case 1: return "cx";
    case 2: return "dx";
    case 3: return "bx";
    case 4: return "sp";
    case 5: return "bp";
    case 6: return "si";
    case 7: return "di";
    default: abort();
    }
}

static char* decode_segreg(byte bits2) {
    switch (bits2 & 0x3) {
    case 0: return "es";
    case 1: return "cs";
    case 2: return "ss";
    case 3: return "ds";
    default: abort();
    }
}

static char* decode_reg8(byte bits3) {
    switch (bits3 & 0x7) {
    case 0: return "al";
    case 1: return "cl";
    case 2: return "dl";
    case 3: return "bl";
    case 4: return "ah";
    case 5: return "ch";
    case 6: return "dh";
    case 7: return "bh";
    default: abort();
    }
}

static char* decode_mem(byte bits3) {
    switch (bits3 & 0x7) {
    case 0: return "bx + si";
    case 1: return "bx + di";
    case 2: return "bp + si";
    case 3: return "bp + di";
    case 4: return "si";
    case 5: return "di";
    case 6: return "bp";
    case 7: return "bx";
    default: abort();
    }
}

static const char* get_size_qualifier(struct a86d_ins* ins, byte opind) {
    if (ins->optype[opind] != a86d_oper_directaddr &&
        ins->optype[opind] != a86d_oper_mem) {
        return "";
    }
    if (ins->optype[0] == a86d_oper_reg
        || ins->optype[1] == a86d_oper_reg
        || ins->optype[0] == a86d_oper_segreg
        || ins->optype[1] == a86d_oper_segreg) {
        /* qualifier not necessary */
        return "";
    }
    switch (ins->opsize[opind]) {
    case 8: return "byte ";
    case 16: return "word ";
    /* nothing necessary since far pointers are unambiguous. */
    case 32: return "";
    default: abort(); /* invalid */
    }
}

static const char* get_seg_prefix(int s) {
    switch (s) {
    case a86d_seg_es: return "es:";
    case a86d_seg_cs: return "cs:";
    case a86d_seg_ss: return "ss:";
    case a86d_seg_ds: return "ds:";
    case a86d_seg_none: return "";
    default: abort();
    }
}

static const char* get_rep_prefix(struct a86d_ins* ins) {
    if (ins->rep_mode == a86d_no_rep) {
        return "";
    }
    if (ins->op == a86d_op_scasw
        || ins->op == a86d_op_scasb
        || ins->op == a86d_op_cmpsw
        || ins->op == a86d_op_cmpsw) {
        if (ins->rep_mode == a86d_repne) {
            return "repne ";
        } else {
            return "repe ";
        }
    } else {
        return "rep ";
    }
}

static const char* get_lock_prefix(struct a86d_ins* ins) {
    if (ins->lock) {
        return "lock ";
    }
    return "";
}

static const char* get_mnemonic(struct a86d_ins* ins) {
    return mnemonic_table[ins->op];
}

static const char* get_reg_name(struct a86d_ins* ins, byte opind) {
    if (ins->opsize[opind] == 8) {
        return decode_reg8(ins->operand[opind]);
    } else { /* 16 assumed. */
        return decode_reg16(ins->operand[opind]);
    }
}

/* statically allocated buffer used for result. */
static const char* get_immediate_hex(struct a86d_ins* ins, byte opind) {
    static char scratch[16];
    scratch[0] = '\0';
    switch (ins->opsize[opind]) {
    case 8:
        snprintf(scratch, 15, "0x%02hhx", ins->operand[opind] & 0xff);
        break;
    case 16:
        snprintf(scratch, 15, "0x%04hx", ins->operand[opind] & 0xffff);
        break;
    case 32:
        snprintf(scratch, 15, "0x%04hx:0x%04hx",
               (ins->operand[opind] & 0xffff0000) >> 16,
               (ins->operand[opind] & 0x0000ffff));
        break;
    default:
        abort();
    }
    return scratch;
}

static const char* get_memref(struct a86d_ins* ins, byte opind) {
    static char scratch[32];
    char scratch2[8] = {0};
    scratch[0] = '[';
    scratch[1] = '\0';
    strcat(scratch, get_seg_prefix(ins->seg));
    strcat(scratch, decode_mem(ins->operand[opind]));
    if (ins->dispsize) {
        strcat(scratch, " + ");
        snprintf(scratch2, 7, "0x%04hx", (word)(ins->disp & 0xffff));
        strcat(scratch, scratch2);
    }
    strcat(scratch, "]");
    return scratch;
}

static const char* get_directref(struct a86d_ins* ins, byte opind) {
    static char scratch[24];
    char scratch2[16] = {0};
    scratch[0] = '[';
    scratch[1] = '\0';
    strcat(scratch, get_seg_prefix(ins->seg));
    snprintf(scratch2, 15, "0x%04hx]", (word)(ins->operand[opind] & 0xffff));
    strcat(scratch, scratch2);
    return scratch;
}


/* disassemble a single decoded instruction. */
char* a86d_disasm(struct a86d_ins* ins, char* outbuf) {
    dword i;
    outbuf[0] = '\0';
    if (ins->op == a86d_op_none || ins->op == a86d_op_db) {
        strcat(outbuf, get_mnemonic(ins));
        return outbuf;
    }
    strcat(outbuf, get_rep_prefix(ins));
    strcat(outbuf, get_lock_prefix(ins));
    strcat(outbuf, get_mnemonic(ins));
    for (i = 0; i < 2; i++) {
        if (i > 0 && ins->optype[i] != a86d_oper_none) {
            strcat(outbuf, ",");
        }
        strcat(outbuf, (" "));

        /* size qualifier if necessary */
        strcat(outbuf, get_size_qualifier(ins, i));

        switch(ins->optype[i]) {
        case a86d_oper_none:
            return outbuf;
        case a86d_oper_reg:
            strcat(outbuf, get_reg_name(ins, i));
            break;
        case a86d_oper_segreg:
            strcat(outbuf, decode_segreg(ins->operand[i]));
            break;
        case a86d_oper_imm:
        case a86d_oper_relimm:
            strcat(outbuf, get_immediate_hex(ins, i));
            break;
        case a86d_oper_mem:
            strcat(outbuf, get_memref(ins, i));
            break;
        case a86d_oper_directaddr:
            strcat(outbuf, get_directref(ins, i));
            break;
        }
    }
    return outbuf;
}
