/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024. Public domain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bits.h"
#include "diskette.h"
#include "memory.h"

#define MAX_DISKETTE_SIZE (1440UL * 1024UL)
#define SECTOR_SIZE 512

void copy_from_addr(byte* to, word from_seg, word from_off, dword count) {
    while (count) {
        *to = load8(from_seg, from_off);
        to++;
        from_off++;
        count--;
    }
}

void copy_to_addr(word to_seg, word to_off, byte* from, dword count) {
    while (count) {
        store8(to_seg, to_off, *from);
        from++;
        to_off++;
        count--;
    }
}

struct diskette_geom {
    dword bytes;
    byte cylinders;
    byte heads;
    byte sectors_per_track;
};

enum diskette_drive_status {
    DSK_NO_DRIVE = 0, /* no drive present */
    DSK_NO_DISK = 1,  /* drive is present, no disk inserted */
    DSK_RO = 2,       /* disk inserted, write-protected */
    DSK_CLEAN = 3     /* disk inserted, writable, no pending changes */
/* disk inserted, writable, changes pending */
/*    DSK_DIRTY = 4;   */
};

static struct {
    byte mem[MAX_DISKETTE_SIZE];
    enum diskette_drive_status status;
    dword size;
} d[DISKETTE_MAX];

static struct diskette_geom geom_map[] = {
    {1440 * 1024, 80, 2, 18},
    {1200 * 1024, 80, 2, 15},
    {720 * 1024, 80, 2, 9},
    {360 * 1024, 40, 2, 9},
    {320 * 1024, 40, 2, 8},
    {180 * 1024, 40, 1, 9},
    {160 * 1024, 40, 1, 8},
    {0,0,0,0}
};

static struct diskette_geom* get_geom(dword bytes) {
    int i = 0;
    while (geom_map[i].bytes) {
        if (geom_map[i].bytes == bytes) {
            return &geom_map[i];
        }
        i++;
    }
    return NULL;
}

/*
cylinder, head, sector are all zero-based here.
in BIOS the sector is one-based
so this normally should be called with (sector - 1).
returns the lba on success
returns -1 on failure
*/
static int32_t chs_to_lba(byte cylinder, byte head, byte sector, byte count, byte drive) {
    int32_t lba = 0;
    /* must correspond to an actual inserted disk */
    struct diskette_geom* geom = get_geom(d[drive].size);
    if (!geom) {
        return -1;
    }
    /* sanity check, drive must exist and contain a disk. */
    if (d[drive].status == DSK_NO_DRIVE || d[drive].status == DSK_NO_DISK) {
        return -1;
    }
    if (cylinder >= geom->cylinders ||
        head >= geom->heads ||
        sector >= geom->sectors_per_track) {
        return -1;
    }
    lba += cylinder * geom->sectors_per_track * geom->heads;
    lba += geom->sectors_per_track * head;
    lba += sector;
    /*
    lba plus count shall not address beyond the last sector of the actual disk.
    TODO should this be greater-or-equal? I think so.
    TODO is this the correct behavior?
    */
    if (lba + count >= (int32_t)(d[drive].size / SECTOR_SIZE)) {
        return -1;
    }
    return lba;
}

/* returns 0 on success, -1 on failure */
int diskette_insert(char* path, byte drive, byte ro) {
    FILE* f = NULL;
    dword bytesread = 0;
    struct diskette_geom* geom;
    int exit_status = -1;

    if (drive >= DISKETTE_MAX) {
        printf("Invalid drive number %hhu\n", drive);
        goto insert_end;
    }
    if (d[drive].status == DSK_NO_DRIVE) {
        printf("No drive %hhu is present\n", drive);
        goto insert_end;
    }
    if (!(f = fopen(path, "rb"))) {
        fprintf(stderr, "%s: ", path);
        perror("Failed to open disk image");
        goto insert_end;
    }
    bytesread = fread(d[drive].mem, 1, MAX_DISKETTE_SIZE, f);
    /* TODO check for disk image larger than max size, and disallow. */
    if (ferror(f)) {
        fprintf(stderr, "Error reading from %s: ", path);
        perror(NULL);
        goto insert_end;
    }
    if (!(geom = get_geom(bytesread))) {
        fprintf(stderr, "%s: Invalid disk size of %u bytes\n", path, bytesread);
        goto insert_end;
    }
    if (ro) {
        d[drive].status = DSK_RO;
    } else {
        d[drive].status = DSK_CLEAN;
    }
    d[drive].size = bytesread;
    exit_status = 0;
insert_end:
    if (f) {
        fclose(f);
    }
    return exit_status;
}

int diskette_reset(byte drive) {
    /* stub. TODO does this need to do anything? */
    (void)drive;
    return 0;
}

int diskette_read(byte cylinder, byte head, byte sector, byte count,
                  byte drive, word to_seg, word to_off) {
    /* TODO actually good implementation. */
    byte* from;
    int32_t lba = 0;
    /* adjust one-based sector numbering used by bios */
    sector -= 1;
    if ((lba = chs_to_lba(cylinder, head, sector, count, drive)) < 0) {
        return -1;
    }
    if (!count) {
        /* read zero is always successful. TODO is this correct? */
        return 0;
    }
    from = d[drive].mem + (lba * 512);
    copy_to_addr(to_seg, to_off, from, count * 512);
    return count;
}

int diskette_write(byte cylinder, byte head, byte sector, byte count,
                   byte drive, word from_seg, word from_off) {
    byte* to;
    int32_t lba = 0;
    /* adjust one-based sector numbering used by bios */
    sector -= 1;
    if ((lba = chs_to_lba(cylinder, head, sector, count, drive)) < 0) {
        return -1;
    }
    if (!count) {
        /* read zero is always successful. TODO is this correct? */
        return 0;
    }
    to = d[drive].mem + (lba * 512);
    copy_from_addr(to, from_seg, from_off, count * 512);
    return count;
}

int diskette_format(byte cylinder, byte head, byte sector, byte count,
                    byte drive, word buf_seg, word buf_off) {
    /* TODO stub */
    (void)buf_seg, (void)buf_off;
    /* adjust one-based sector numbering used by bios */
    sector -= 1;
    if (chs_to_lba(cylinder, head, sector, count, drive) < 0) {
        return -1;
    }
    if (!count) {
        return 0;
    }
    return -1;
}

int diskette_verify(byte cylinder, byte head, byte sector, byte count,
                    byte drive) {
    /* TODO stub */
    if (chs_to_lba(cylinder, head, sector, count, drive) < 0) {
        return -1;
    }
    return count;
}

void diskette_init(byte drives) {
    unsigned int i;
    if (drives > DISKETTE_MAX) {
        fprintf(stderr,
                "Number of drives must be in the range [0,%u]. Not %hhu\n",
                DISKETTE_MAX,
                drives);
        exit(1);
    }
    for (i = 0; i < DISKETTE_MAX; i++) {
        memset(d[i].mem, 0, MAX_DISKETTE_SIZE);
        if (i < drives) {
            d[i].status = DSK_NO_DISK;
        } else {
            d[i].status = DSK_NO_DRIVE;
        }
        d[i].size = 0;
    }
}

void diskette_dump_state(void) {
    unsigned int i = 0;
    puts("+------+---------------+---------------+---------------+---------------+");
    printf("| fd   ");
    for (i = 0; i < DISKETTE_MAX; i++) {
        printf("| %u: ", i);
        switch (d[i].status) {
        case DSK_NO_DRIVE:
            printf("no drive   ");
            break;
        case DSK_NO_DISK:
            printf("empty      ");
            break;
        case DSK_RO:
            printf("%4uk  [R] ", d[i].size / 1024);
            break;
        case DSK_CLEAN:
            printf("%4uk  [ ] ", d[i].size / 1024);
            break;
        default:
            abort();
        }
    }
    printf("|\n");
    puts("+------+---------------+---------------+---------------+---------------+");
}

void diskette_eject(byte drive) {
    if (drive >= DISKETTE_MAX) {
        printf("Invalid drive number %hhu\n", drive);
    } else if (d[drive].status == DSK_NO_DISK) {
        printf("No disk in drive %hhu\n", drive);
    } else if (d[drive].status == DSK_NO_DRIVE) {
        printf("No drive %hhu is present\n", drive);
    } else {
        d[drive].size = 0;
        d[drive].status = DSK_NO_DISK;
        printf("Removed disk in drive %hhu\n", drive);
    }
}
