/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bits.h"
#include "cpu.h"
#include "debug.h"
#include "diskette.h"
#include "f600.h"
#include "memory.h"
#include "pc.h"
#include "str.h"
#include "term.h"

#define shift() (argc--, argv++)

/* get option argument */
#define optarg() (\
    (argv[0][1]) ?\
        (nextarg = 1, &argv[0][1]) : (\
        (argc > 1) ?\
            (nextarg = 1, shift(), argv_[1]) :\
            (usage(), abort(), (char*)NULL)\
        )\
)\

char* argv0;

void die(const char* s, ...) {
    va_list ap;
    int e = errno;
    fprintf(stderr, "%s: ", argv0);
    va_start(ap, s);
    vfprintf(stderr, s, ap);
    va_end(ap);
    if (e) {
        fprintf(stderr, ": %s", strerror(e));
    }
    fputc('\n', stderr);
    fflush(stderr);
    exit(1);
}

void usage(void) {
    fprintf(stderr,
"usage: %s [-d n] [-b path] bios-img [disk-img]...\n"
"  -b path  path to 32KiB BASIC ROM\n"
"  -d n     number of diskette drives (0, 1 or 2. default = 2)\n"
"  -s       start with emulated PC turned off\n"
, argv0);
    exit(1);
}

byte* get_file_contents(char* path, size_t expected_size) {
    byte* buf;
    size_t actual_size = 0;
    FILE* f = fopen(path, "r");
    buf = malloc(expected_size);
    if (!buf) {
        fprintf(stderr, "malloc failed\n");
        exit(1);
    }
    if (!f) {
        perror(path);
        exit(1);
    }
    while (fgetc(f) != EOF) {
        actual_size++;
    }
    if (ferror(f)) {
        perror("Read error");
        exit(1);
    }
    if (actual_size != expected_size) {
        fprintf(stderr, "Error: file must be %lu bytes, not %lu\n",
                expected_size, actual_size);
        exit(1);
    }
    rewind(f);
    if ((actual_size = fread(buf, 1, expected_size, f)) != expected_size) {
        fprintf(stderr, "Read error, expected %lu bytes, got %lu\n",
                expected_size, actual_size);
        perror(NULL);
        exit(1);
    }
    return buf;
}

void load_roms(char* bios_path, char* basic_path) {
    byte* buf = get_file_contents(bios_path, 8192);
    load_bios(buf);
    free(buf);
    if (!basic_path) {
        buf = malloc(32768);
        if (!buf) {
            fprintf(stderr, "malloc failed\n");
            exit(1);
        }
        memset(buf, 0, 32768);
        memcpy(buf, f600_default_rom, F600_DEFAULT_ROM_LEN);
    } else {
        buf = get_file_contents(basic_path, 32768);
    }
    load_basic(buf);
    free(buf);
}

int main(int argc, char* argv[]) {
    int i = 0;
    int nextarg = 0;
    char** argv_;
    char* basic_path = NULL;
    byte drives = 2;
    byte start_off = 0;

    argv0 = argv[0];
    shift();
    for (; argc && argv[0][0] == '-' && argv[0][1]; shift()) {
        if (argv[0][1] == '-' && !argv[0][2]) {
            shift();
            break;
        }
        nextarg = 0;
        for (argv[0]++; !nextarg && argv[0][0]; argv[0]++) {
            argv_ = argv;
            switch (argv[0][0]) {
            case 'd':
                if (strtou8(optarg(), &drives) < 0 || drives > 2) {
                    die("argument to -d must be one of: 0, 1, 2");
                }
                break;
            case 's':
                start_off = 1;
                break;
            case 'b':
                basic_path = optarg();
                break;
            default:
                usage();
            }
        }
    }

    if (argc < 1 || argc > (1 + drives)) {
        usage();
    }
    pc_init((byte)drives);
    load_roms(argv[0], basic_path);
    shift();
    while (argc) {
        /* TODO allow setting RO flag */
        if (diskette_insert(*argv, i++, 0) < 0) {
            die("Disk image load failed");
        }
        shift();
    }
    term_init();
    atexit(term_reset);
    term_set_raw();
    pc_run(debug_cmd, start_off);
    return 0;
}
