/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bits.h"
#include "cp437.h"
#include "term.h"   /* for escape codes */

/* IBM Monochrome Display Adapter */

#define rows 25
#define vram_size 4096

/* arbitrary value big enough to hold all attributes, etc */
#define OUTBUFSIZE 16384

#define pages ((mode_control & 1) ? 1 : 2)
#define cols ((mode_control & 1) ? 80 : 40)
#define video_enabled (!!(mode_control & 8))
#define blink_enabled (!!(mode_control & 32))

static byte* vram;
static byte prev_vram[vram_size];

static byte page;

static char outbuf[OUTBUFSIZE];

static byte prev_cur_row = 0;
static byte prev_cur_col = 0;

static byte mda_status = 0;

/*
mode control register bits:
0: 1 = 80x25 text, 0 = 40x25 text
1: unused
2: unused
3: 1 = enable video signal
4: unused
5: blinking on
6: unused
7: unused
*/
static byte mode_control = 0;
static byte prev_mode_control = 0;

#define CHAR(row, col) (vram[((row) * cols + (col)) * 2])
#define ATTR(row, col) (vram[((row) * cols + (col)) * 2 + 1])

/*
6845 controller registers
12 is msb of start address
13 is lsb of start address
14 is msb of cursor address
15 is lsb of cursor address
*/
static byte reg[16] = {0};
static byte which_reg = 0;

static void cur_address_to_curxy(byte* row, byte* col) {
    word curaddr = (word)(reg[14] << 8) | (word)(reg[15]);
    *row = (curaddr / 80) % 25; /* the mod 25 is a safety. */
    *col = (curaddr % 80);
}

static inline byte vram_needs_refresh(void) {
    return memcmp(vram, prev_vram, vram_size) || mode_control != prev_mode_control;
}

static inline byte cursor_needs_refresh(void) {
    byte cur_row = 0;
    byte cur_col = 0;
    cur_address_to_curxy(&cur_row, &cur_col);
    return cur_row != prev_cur_row || cur_col != prev_cur_col;
}

void vram_refresh(void) {
    int row = 0;
    int col = 0;
    char attr = 0;
    char character[2] = {0};
    byte attrib = 0;
    byte intensity = 0;
    byte blink = 0;
    *outbuf = 0;
    strcat(outbuf, ESC_RESET_CURSOR);
    prev_mode_control = mode_control;
    if (!video_enabled) {
        /* blank */
        strcat(outbuf, ESC_CLEAR);
        return;
    }
    /* TODO use start address from registers */
    for (row = 0; row < rows; row++) {
        for (col = 0; col < cols; col++) {
            character[0] = CHAR(row, col);
            attr = ATTR(row, col);
            attrib = attr & 0x77; /* excludes blink and intensity */
            intensity = attr & 0x8;
            blink = attr & 0x80;
            switch (attrib) {
            case 0:
                /* hidden */
                strcat(outbuf, " ");
                continue;
            case 1:
                /* underline */
                strcat(outbuf, ESC_UNDERLINE);
                break;
            /*case 0:*/
            case 7:
                /* normal */
                break;
            case 0x70:
                /* reverse video */
                strcat(outbuf, ESC_REVERSE_VIDEO); /* reverse video */
                break;
            default:
                /* ignore junk TODO is this the correct behavior? */
                break;
            }
            if (blink && blink_enabled) {
                strcat(outbuf, ESC_BLINK);
            }
            if (intensity) {
                strcat(outbuf, ESC_BOLD);
            }
            strcat(outbuf, cp437_to_utf8[(byte)(character[0])]);
            strcat(outbuf, ESC_RESET_ATTRS);
        }
        strcat(outbuf, "\n");
    }
}

/* vram: pointer to MDA video ram area at address B000:0000 */
void mda_refresh(int force) {
    byte cur_row = 0;
    byte cur_col = 0;
    byte do_vram_refresh = force || vram_needs_refresh();
    byte do_cursor_refresh = do_vram_refresh || cursor_needs_refresh();
    if (!do_vram_refresh && !do_cursor_refresh) {
        return;
    }
    if (do_vram_refresh) {
        vram_refresh();
        printf("%s", outbuf);
        memcpy(prev_vram, vram, vram_size);
    }
    if (do_cursor_refresh) {
        cur_address_to_curxy(&cur_row, &cur_col);
        term_set_cursor(cur_row, cur_col);
        prev_cur_row = cur_row;
        prev_cur_col = cur_col;
    }
    fflush(stdout);
}

void mda_init(byte* vram_ptr) {
    vram = vram_ptr;
    page = 0;
    memset(outbuf, 0, OUTBUFSIZE);
    memset(vram, 0, vram_size);
    memset(prev_vram, 0, vram_size);
    memset(reg, 0, 16);
    prev_cur_row = 0;
    prev_cur_col = 0;
    mode_control = 1; /* start in 80x25 */
    prev_mode_control = 0;
    /* status bits (from documentation:
       7-4 unused
       3   vertical retrace, ram access ok for next 1.25 ms
       2   0 = light pen on, 1 = light pen off
       1   1 = light pen trigger set
       0   1 = display enable, ram access is ok
    */
    mda_status = 4;
    printf(ESC_CLEAR ESC_RESET_CURSOR);
    fflush(stdout);
}

void mda_set_register_value(byte val) {
    reg[which_reg] = val;
}

void mda_select_register(byte which) {
    which_reg = which & 0xf;
    /* TODO what about the light pen, supposedly readable
       from registers 0x10, 0x11? */
}

byte mda_get_register_value(void) {
    return reg[which_reg];
}

/*
int main(void) {
    byte buf[OUTBUFSIZE];
    mda_init(buf);
    memset(buf, 'p', OUTBUFSIZE);
    mda_refresh();
    return 0;
}
*/

void mda_dump_state(void) {
    int i;
    puts("+------------------------------------------------------+---------------+");
    printf("| mda 00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  | row %2i        |\n",
    prev_cur_row);
    printf("|     ");
    for (i = 0; i < 16; i++) {
        printf("%02hhx ", reg[i]);
    }
    printf(" | col %2i        |\n", prev_cur_col);
}

byte mda_get_status(void) {
    /* flip bit 0 and 3 each time this is accessed, I am hoping this is a good
       enough simulation for now.
       ignoring light pen stuff. light pen is always off (bit 2 is set)
    */
    if (mda_status == 13) {
        mda_status = 4;
    } else {
        mda_status = 13;
    }
    return mda_status;
}

void mda_set_mode_control(byte b) {
    prev_mode_control = mode_control;
    mode_control = b;
}

byte mda_get_mode_control(void) {
    return mode_control;
}
