/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/


#include <string.h>
#include <stdio.h>

#include "bits.h"

static byte mem[0xa0000];
static byte mda_mem[0x4000]; /* mapped to b000:0000 */
static byte rom[0xa000];

/*
simulate 8-bit bus so that when writing/reading invalid location
a "phantom" value will be returned.
TODO 16-bit bus as in the 8086.
*/
static byte bus;

static void load_to_bus(word seg, word off) {
    dword flat = segoff_to_flat(seg, off);
    if (flat >= 0xb0000 && flat <= 0xb7fff) {
        /* 4 kb MDA area addressable at b000:0000, b100:0000, b200:0000 ...
           through b700:0000. */
        flat &= ~(dword)(0x7000);
        bus = mda_mem[flat - 0xb0000];
    } else if (flat >= 0xf6000) {
        /* ROM */
        bus = rom[flat - 0xf6000];
    } else if (flat < 0xa0000) {
        /* conventional memory */
        bus = mem[flat];
    } else {
        /* nothing */
    }
}

static void store_from_bus(word seg, word off) {
    dword flat = segoff_to_flat(seg, off);
    if (flat >= 0xb0000 && flat <= 0xb7fff) {
        /* MDA area */
        flat &= ~(dword)(0x7000);
        mda_mem[flat - 0xb0000] = bus;
    } else if (flat < 0xa0000) {
        /* conventional memory */
        mem[flat] = bus;
    } else {
        /* ROM or nothing */
    }
}

byte load8(word seg, word off) {
    load_to_bus(seg, off);
    return bus;
}

word load16(word seg, word off) {
    /* wraps around at offset 0xffff, does not fault. */
    word result = 0;
    result |= load8(seg, off);
    result |= (word)(load8(seg, off + 1)) << 8;
    return result;
}

dword load32(word seg, word off) {
    dword result = 0;
    result |= load8(seg, off);
    result |= (dword)(load8(seg, off + 1)) << 8;
    result |= (dword)(load8(seg, off + 2)) << 16;
    result |= (dword)(load8(seg, off + 3)) << 24;
    return result;
}

void store8(word seg, word off, byte val) {
    bus = val;
    store_from_bus(seg, off);
}

void store16(word seg, word off, word val) {
    store8(seg, off, LB(val));
    store8(seg, off + 1, HB(val));
}

void store32(word seg, word off, dword val) {
    store8(seg, off, LB(val));
    store8(seg, off + 1, HB(val));
    store8(seg, off + 2, LB(HW(val)));
    store8(seg, off + 3, HB(HW(val)));
}

byte* memory_mda_vram(void) {
    return mda_mem;
}

void load_basic(byte* from) {
    memcpy(rom, from, 32768);
}

void load_bios(byte* from) {
    memcpy(rom + 32768, from, 8192);
}

void memory_dump_at(word seg, word off) {
    byte i, j;
    printf("           00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f\n");
    printf("----------------------------------------------------------\n");
    int c;
    for (i = 0; i < 16; i++) {
        printf("%04hx:%04hx  ", seg, off);
        for (j = 0; j < 16; j++) {
            c = load8(seg, off);
            printf("%02hhx ", c);
            off++;
        }
        off -= 16;
        printf("  ");
        for (j = 0; j < 16; j++) {
            c = load8(seg, off);
            printf("%c", (c < 0x20 || c >= 0x7f) ? '.' : c);
            off++;
        }
        printf("\n");
    }
}
