/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

#include <stdio.h>
#include <stdlib.h>

#include "bits.h"
#include "breakpoint.h"
#include "clock.h"
#include "cpu.h"
#include "diskette.h"
#include "dma.h"
#include "mda.h"
#include "memory.h"
#include "periph.h"
#include "pic.h"
#include "term.h"
#include "timer.h"

/*
Poll keyboard and update screen every 2^n emulated cycles.
(16 = 65536 cycles)
*/
#define IO_REFRESH_INTERVAL 16
    
static byte nmi_enabled = 1;
int debug_flag = 0;
static byte diskette_drives = 0;

static dword pc_io_in(word port) {
    switch(port) {
    case 0x00: return dma_read_reg(0);
    case 0x01: return dma_read_reg(1);
    case 0x02: return dma_read_reg(2);
    case 0x03: return dma_read_reg(3);
    case 0x04: return dma_read_reg(4);
    case 0x05: return dma_read_reg(5);
    case 0x06: return dma_read_reg(6);
    case 0x07: return dma_read_reg(7);
    case 0x21: return pic_get_mask(); /* TODO tmp hack for PIC interrupt masking */
    case 0x40: return timer_read(0);
    case 0x41: return timer_read(1);
    case 0x42: return timer_read(2);
    case 0x60: return periph_port_a_in();
    case 0x61: return periph_port_b_in(); break;
    case 0x62: return periph_port_c_in();
    case 0x3b1:
    case 0x3b3:
    case 0x3b5:
    case 0x3b7:
    /* mda data in */
        return mda_get_register_value();
    case 0x3b8:
        return mda_get_mode_control();
    case 0x3ba:
    /* mda status */
        return mda_get_status();
    default: return UINT32_MAX;
    }
}

static void pc_io_out(word port, word data) {
    switch(port) {
    case 0x00: dma_write_reg(0, (byte)data); break;
    case 0x01: dma_write_reg(1, (byte)data); break;
    case 0x02: dma_write_reg(2, (byte)data); break;
    case 0x03: dma_write_reg(3, (byte)data); break;
    case 0x04: dma_write_reg(4, (byte)data); break;
    case 0x05: dma_write_reg(5, (byte)data); break;
    case 0x06: dma_write_reg(6, (byte)data); break;
    case 0x07: dma_write_reg(7, (byte)data); break;
    case 0x21: pic_set_mask(data & 0xff); break; /* TODO tmp hack for PIC interrupt masking */
    case 0x40: timer_set_maxval(0, (byte)data); break;
    case 0x41: timer_set_maxval(1, (byte)data); break;
    case 0x42: timer_set_maxval(2, (byte)data); break;
    case 0x43: timer_set_mode((byte)data); break;
    case 0x61: periph_port_b_out((byte)data); break;
    case 0xa0: nmi_enabled = !!(data & 0x80); break;  /* NMI mask port */
    case 0x3b0:
    case 0x3b2:
    case 0x3b4:
    case 0x3b6:
    /* mda register select */
        mda_select_register((byte)data);
        break;
    case 0x3b1:
    case 0x3b3:
    case 0x3b5:
    case 0x3b7:
    /* mda data out */
        mda_set_register_value((byte)data);
        break;
    case 0x3b8:
        mda_set_mode_control((byte)data);
        break;
    default: break;
    }
}

static byte irq_dummy(void) {
    return 0;
}

static int keyfunc(void) {
    int k = term_next_key();
    if (k == MAGIC_DEBUG_KEY) {
        debug_flag = 1;
        return -1;
    }
    return k;
}
    
static byte (*pc_irq_router[8])(void) = {
    timer_poll_timer_0, /* IRQ 0 */
    periph_poll_irq1,   /* IRQ 1 */
    irq_dummy,          /* IRQ 2 */
    irq_dummy,          /* IRQ 3 */
    irq_dummy,          /* IRQ 4 */
    irq_dummy,          /* IRQ 5 */
    irq_dummy,          /* IRQ 6 */
    irq_dummy           /* IRQ 7 */
};

void pc_run(int (*debug_func)(void), byte start_off) {
    uint64_t io_interval = (1 << IO_REFRESH_INTERVAL) - 1;
    int bp = 0;
    cpu_reset();
    cpu_set_io_in_func(pc_io_in);
    cpu_set_io_out_func(pc_io_out);
    dma_init();
    timer_init();
    pic_init(pc_irq_router);
    periph_init(keyfunc, diskette_drives);
    mda_init(memory_mda_vram());
    if (start_off) {
        debug_flag = 1;
    }
    while (1) {
        if (breakpoints) {
            if ((bp = breakpoint_hit(cpu_get_cs(), cpu_get_ip())) >= 0) {
                printf("breakpoint %i hit at %04hx:%04hx\n", bp, cpu_get_cs(), cpu_get_ip());
                debug_flag = 1;
            }
        }
        cpu_fetch();
        if (debug_flag) {
            debug_flag = 0;
            if (!debug_func) {
                fprintf(stderr, "No debug function specified");
                exit(1);
            }
            while (debug_func());
        }
        cpu_execute();
        if (!(clock_get_cycle_count() & io_interval)) {
            periph_pollkey();
            mda_refresh(0);
        }
    }
}

void pc_init(byte drives) {
    diskette_init(drives);
    diskette_drives = drives;
}
