/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

/* Intel 8255A peripheral chip */

#include <stdio.h>
#include <stdlib.h>

#include "bits.h"
#include "periph.h"
#include "clock.h"

static byte timer_2_gate_speaker = 0;
static byte read_rw_mem_size = 0;
static byte cassette_motor_off = 0;
static byte enable_rw_mem = 0;
static byte enable_io_channel_check = 0;
static byte hold_kb_clock_low = 0;
static byte keyboard_enable = 0;
static byte do_reset = 0;
static byte clock_hold_start = 0;
static byte porta = 0; /* raw value */
static byte portb = 0; /* raw value */
static byte portc = 0; /* raw value */
static byte drives = 0; /* number of diskette drives */

static byte irq1 = 0; /* keyboard irq */
static byte scancode = 0; /* keyboard scancode */

static int dummy(void) {
    return -1;
}

static int (*kfunc)(void) = dummy;

/* TODO care about keyboard clock line being held low by bios? */

byte periph_port_a_in(void) {
    byte result = 0;
    if (keyboard_enable) {
        result = scancode;
    } else {
        result |= ((drives > 0) ? 1 : 0); /* IPL 5-1/4 diskette drive is present */
        result |= 4 | 8; /* 64k base ram on board */
        result |= 16 | 32; /* 80x25 monochrome mode on power up */
        switch (drives) {
        case 0:
        case 1:
            break;
        case 2:
            result |= 64;
            break;
        case 3:
            result |= 128;
            break;
        case 4:
            result |= 64;
            result |= 128;
            break;
        default:
            abort();
        }
    }
    porta = result;
    return result;
}

void periph_port_b_out(byte inbyte) {
    portb = inbyte;
    timer_2_gate_speaker = !!(inbyte & 1);
    /* bit 2 is speaker data, TODO use this. */
    read_rw_mem_size = !!(inbyte & 4);
    cassette_motor_off = !!(inbyte & 8);
    enable_rw_mem = !(inbyte & 16);
    enable_io_channel_check = !(inbyte & 32);
    if (!(inbyte & 128)) {
        /* keyboard enable */
        if (!keyboard_enable) {
            if (do_reset) {
                irq1 = 1;
                do_reset = 0;
                scancode = 0xAA; /* reset scancode */
            }
        }
        keyboard_enable = 1;
    } else {
        keyboard_enable = 0;
    }
    if (!(inbyte & 64)) { /* set low */
        if (!hold_kb_clock_low) {
            hold_kb_clock_low = 1;
            clock_hold_start = clock_get_cycle_count();
        }
    } else { /* set high */
        if (hold_kb_clock_low) {
            if (clock_get_cycle_count() - clock_hold_start > 10000) {
                do_reset = 1;
            }
            hold_kb_clock_low = 0;
        }
    }
}

byte periph_port_b_in(void) {
    return portb;
}

byte periph_port_c_in(void) {
    /* TODO impl */
    byte result = 0;
    /* total 640 k RAM */
    if (read_rw_mem_size) {
        result |= 0x02; /* multiple of 32k */
    } else {
       result |= 0x01; /* multiple of 16 x 32k */
    }
    portc = result;
    return result;
}

byte periph_poll_irq1(void) {
    if (irq1) {
        irq1 = 0;
        return 1;
    }
    return 0;
}

void periph_dump_state(void) {
    puts("+---------------------------------------+-A--+-B--+-C--+---------------+");
    printf("| kb enabled: [%c] scancode: [%02hhx]        | %02hhx | %02hhx | %02hhx | irq 1: [%c]    |\n",
        keyboard_enable ? 'x' : 'x', scancode, porta, portb, portc, irq1 ? 'x' : ' ');
}

void periph_pollkey(void) {
    int c = kfunc();
    if (!keyboard_enable) {
        return;
    }
    if (c != -1) {
        scancode = c & 0xff;
        irq1 = 1;
    }
}

void periph_init(int (*keyfunc)(void), byte diskette_drives) {
    if (keyfunc) {
        kfunc = keyfunc;
    }
    if (diskette_drives > 4) {
        fprintf(stderr,
                "Number of drives must be in the range [0,4]. Not %hhu\n",
                drives);
        exit(1);
    }
    drives = diskette_drives;
}
