/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022-2024.
Public domain.
*/

/* POSIX terminal and keyboard handling */

#define _POSIX_C_SOURCE 200809L

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <poll.h>
#include <unistd.h>

#include "term.h"

#define KB_QUEUE_MAX 32
#define ALT_ON 0x38
#define ALT_OFF 0xb8
#define CTRL_ON 0x1d
#define CTRL_OFF 0x9d
#define SHIFT_ON 0x36
#define SHIFT_OFF 0xb6
#define shift(x)    enq(SHIFT_ON); x; enq(SHIFT_OFF)
#define ctrl(x)     enq(CTRL_ON); x; enq(CTRL_OFF)
#define alt(x)     enq(ALT_ON); x; enq(ALT_OFF)
#define SPECIAL_CTRL_KEY 0x1e /* ctrl-6 */
#define SPECIAL_ALT_KEY 0x1d /* ctrl-] */
#define SPECIAL_SHIFT_KEY 0x1f /* ctrl-- */

enum mod_state {
    MOD_NONE = 0,    /* key not being held */
    MOD_PENDING = 1, /* key pressed on host, but key-down event not yet sent */
    MOD_HELD = 2     /* key-down event has been sent */
};

struct mod_key {
    dword host_scancode; /* scancode from host keyboard for this key */
    dword len; /* length of scancode in bytes (1-4) */
    byte on_byte;    /* byte to send indicating key is held */
    byte off_byte;   /* byte to send indicating key is released */
    enum mod_state state; /* current state of key */
};

struct mod_key mod_keys[] = {
    {SPECIAL_ALT_KEY, 1, ALT_ON, ALT_OFF, MOD_NONE},
    {SPECIAL_CTRL_KEY, 1, CTRL_ON, CTRL_OFF, MOD_NONE},
    {SPECIAL_SHIFT_KEY, 1, SHIFT_ON, SHIFT_OFF, MOD_NONE},
    {0, 0, 0, 0, MOD_NONE} /* end of list */
};

static byte kb_queue[KB_QUEUE_MAX] = {0};
static byte kb_queue_count = 0;
static struct termios orig_settings;
static byte inited = 0;

/* enqueue keyboard scan code */
static int enq(byte b) {
    if (kb_queue_count == KB_QUEUE_MAX) {
        return -1; /* full */
    }
    kb_queue[kb_queue_count] = b;
    kb_queue_count++;
    return 0;
}

/* dequeue and return next keyboard scancode, or -1 if none is available. */
static int deq(void) {
    int item = -1;
    int i;
    if (!kb_queue_count) {
        return item; /* empty */
    }
    item = kb_queue[0];
    /* O(n), but fine. */
    for (i = 0; i < kb_queue_count - 1; i++) {
        kb_queue[i] = kb_queue[i + 1];
    }
    kb_queue_count--;
    return item;
}

static int key_available(void) {
    static struct pollfd fd = {.fd = 0, .events = POLLIN};
    if (poll(&fd, 1, 0) <= 0) {
        /* TODO check error */
        return 0;
    }
    if (!(fd.revents & POLLIN)) {
        /* no key available */
        return 0;
    }
    return 1;
}

static void enqueue_single_key(uint64_t result, ssize_t inlen);

static void enqueue_single_key_with_mod(uint64_t result, ssize_t inlen) {
    size_t i;
    for (i = 0; mod_keys[i].len; i++) {
        if (mod_keys[i].len == inlen && mod_keys[i].host_scancode == result) {
            if (mod_keys[i].state == MOD_NONE) {
                mod_keys[i].state = MOD_PENDING;
                return;
            } else {
                /* mod key already pending */
                return;
            }
        }
    }
    for (i = 0; mod_keys[i].len; i++) {
        if (mod_keys[i].state == MOD_PENDING) {
            mod_keys[i].state = MOD_HELD;
            enq(mod_keys[i].on_byte);
        }
    }
    enqueue_single_key(result, inlen);
    for (i = 0; mod_keys[i].len; i++) {
        if (mod_keys[i].state == MOD_HELD) {
            mod_keys[i].state = MOD_NONE;
            enq(mod_keys[i].off_byte);
        }
    }
}

static void enqueue_single_key(uint64_t result, ssize_t inlen) {
    switch (inlen) {
    case 0:
        /* nothing */
        return;
    case 1:
        switch (result) {
        case '\033': enq(0x01); return; /* esc */
        case '1': enq(0x02); return;
        case '2': enq(0x03); return;
        case '3': enq(0x04); return;
        case '4': enq(0x05); return;
        case '5': enq(0x06); return;
        case '6': enq(0x07); return;
        case '7': enq(0x08); return;
        case '8': enq(0x09); return;
        case '9': enq(0x0a); return;
        case '0': enq(0x0b); return;
        case '-': enq(0x0c); return;
        case '=': enq(0x0d); return;
        case '!': shift(enq(0x02)); return;
        case '@': shift(enq(0x03)); return;
        case '#': shift(enq(0x04)); return;
        case '$': shift(enq(0x05)); return;
        case '%': shift(enq(0x06)); return;
        case '^': shift(enq(0x07)); return;
        case '&': shift(enq(0x08)); return;
        case '*': shift(enq(0x09)); return;
        case '(': shift(enq(0x0a)); return;
        case ')': shift(enq(0x0b)); return;
        case '_': shift(enq(0x0c)); return;
        case '+': shift(enq(0x0d)); return;
        case '\x7f': enq(0x0e); return; /* raw backspace key */
        case '\t': enq(0x0f); return; /* tab */
        case 'q': enq(0x10); return;
        case 'w': enq(0x11); return;
        case 'e': enq(0x12); return;
        case 'r': enq(0x13); return;
        case 't': enq(0x14); return;
        case 'y': enq(0x15); return;
        case 'u': enq(0x16); return;
        case 'i': enq(0x17); return;
        case 'o': enq(0x18); return;
        case 'p': enq(0x19); return;
        case '[': enq(0x1a); return;
        case ']': enq(0x1b); return;
        case 'Q': shift(enq(0x10)); return;
        case 'W': shift(enq(0x11)); return;
        case 'E': shift(enq(0x12)); return;
        case 'R': shift(enq(0x13)); return;
        case 'T': shift(enq(0x14)); return;
        case 'Y': shift(enq(0x15)); return;
        case 'U': shift(enq(0x16)); return;
        case 'I': shift(enq(0x17)); return;
        case 'O': shift(enq(0x18)); return;
        case 'P': shift(enq(0x19)); return;
        case '{': shift(enq(0x1a)); return;
        case '}': shift(enq(0x1b)); return;
        case '\r': enq(0x1c); return;
        /* 0x1d left control */
        case 1: ctrl(enq(0x1e)); return; /* ctrl-a */
        case 2: ctrl(enq(0x30)); return; /* ctrl-b */
        case 3: ctrl(enq(0x2e)); return; /* ctrl-c */
        case 4: ctrl(enq(0x20)); return; /* ctrl-d */
        case 5: ctrl(enq(0x12)); return; /* ctrl-e */
        case 6: ctrl(enq(0x21)); return; /* ctrl-f */
        case 7: ctrl(enq(0x22)); return; /* ctrl-g */
        case 8: ctrl(enq(0x23)); return; /* ctrl-h */
        /* ctrl-i is tab */
        /* ctrl-j is newline */
        case 11: ctrl(enq(0x25)); return; /* ctrl-k */
        case 12: ctrl(enq(0x26)); return; /* ctrl-l */
        /* ctrl-m is carriage return */
        case 14: ctrl(enq(0x31)); return; /* ctrl-n */
        case 15: ctrl(enq(0x18)); return; /* ctrl-o */
        case 16: ctrl(enq(0x19)); return; /* ctrl-p */
        case 17: ctrl(enq(0x10)); return; /* ctrl-q */
        case 18: ctrl(enq(0x13)); return; /* ctrl-r */
        case 19: ctrl(enq(0x1f)); return; /* ctrl-s */
        case 20: ctrl(enq(0x14)); return; /* ctrl-t */
        case 21: ctrl(enq(0x16)); return; /* ctrl-u */
        case 22: ctrl(enq(0x2f)); return; /* ctrl-v */
        case 23: ctrl(enq(0x11)); return; /* ctrl-w */
        case 24: ctrl(enq(0x2d)); return; /* ctrl-x */
        case 25: ctrl(enq(0x15)); return; /* ctrl-y */
        case 26: ctrl(enq(0x2c)); return; /* ctrl-z */
        case 'a': enq(0x1e); return;
        case 's': enq(0x1f); return;
        case 'd': enq(0x20); return;
        case 'f': enq(0x21); return;
        case 'g': enq(0x22); return;
        case 'h': enq(0x23); return;
        case 'j': enq(0x24); return;
        case 'k': enq(0x25); return;
        case 'l': enq(0x26); return;
        case ';': enq(0x27); return;
        case '\'': enq(0x28); return;
        case '`': enq(0x29); return;
        case 'A': shift(enq(0x1e)); return;
        case 'S': shift(enq(0x1f)); return;
        case 'D': shift(enq(0x20)); return;
        case 'F': shift(enq(0x21)); return;
        case 'G': shift(enq(0x22)); return;
        case 'H': shift(enq(0x23)); return;
        case 'J': shift(enq(0x24)); return;
        case 'K': shift(enq(0x25)); return;
        case 'L': shift(enq(0x26)); return;
        case ':': shift(enq(0x27)); return;
        case '"': shift(enq(0x28)); return;
        case '~': shift(enq(0x29)); return;
        /* 2a left shift */
        case '\\': enq(0x2b); return;
        case '|': shift(enq(0x2b)); return;
        case 'z': enq(0x2c); return;
        case 'x': enq(0x2d); return;
        case 'c': enq(0x2e); return;
        case 'v': enq(0x2f); return;
        case 'b': enq(0x30); return;
        case 'n': enq(0x31); return;
        case 'm': enq(0x32); return;
        case ',': enq(0x33); return;
        case '.': enq(0x34); return;
        case '/': enq(0x35); return;
        case 'Z': shift(enq(0x2c)); return;
        case 'X': shift(enq(0x2d)); return;
        case 'C': shift(enq(0x2e)); return;
        case 'V': shift(enq(0x2f)); return;
        case 'B': shift(enq(0x30)); return;
        case 'N': shift(enq(0x31)); return;
        case 'M': shift(enq(0x32)); return;
        case '<': shift(enq(0x33)); return;
        case '>': shift(enq(0x34)); return;
        case '?': shift(enq(0x35)); return;
        /* 0x36 right shift */
        case 0x1c: enq(0x37); return; /* print screen */
        /* 0x38 left alt */
        case ' ': enq(0x39); return;
        /* 0x3a caps lock */
        default: return;
        }
    case 2:
        if (result >> 8 == '\033') {
            /* alt+key */
            enq(ALT_ON);
            enqueue_single_key(result & 0xff, 1);
            enq(ALT_OFF);
        }
        return;
    case 3:
        switch (result) {
        case 0x1b5b41: enq(0x48); return; /* arrow up (numpad) */
        case 0x1b5b42: enq(0x50); return; /* arrow down (numpad) */
        case 0x1b5b43: enq(0x4d); return; /* arrow right (numpad) */
        case 0x1b5b44: enq(0x4b); return; /* arrow left (numpad) */
        default: return;
        }
    case 4:
        switch (result) {
        case 0x1b5b317e: enq(0x47); return; /* home */
        case 0x1b5b337e: enq(0x53); return; /* del */
        case 0x1b5b347e: enq(0x4f); return; /* end */
        case 0x1b5b357e: enq(0x49); return; /* pg up */
        case 0x1b5b367e: enq(0x51); return; /* pg down */
        case 0x1b5b5b41: enq(0x3b); return; /* F1 */
        case 0x1b5b5b42: enq(0x3c); return; /* F2 */
        case 0x1b5b5b43: enq(0x3d); return; /* F3 */
        case 0x1b5b5b44: enq(0x3e); return; /* F4 */
        case 0x1b5b5b45: enq(0x3f); return; /* F5 */
        default: return;
        }
    case 5:
        switch (result) {
        case 0x1b5b31377e: enq(0x40); return; /* F6 */
        case 0x1b5b31387e: enq(0x41); return; /* F7 */
        case 0x1b5b31397e: enq(0x42); return; /* F8 */
        case 0x1b5b32307e: enq(0x43); return; /* F9 */
        case 0x1b5b32317e: enq(0x44); return; /* F10 */
        case 0x1b5b32337e: ctrl(enq(0x46)); return; /* F11 as ctrl+break. */
        case 0x1b5b32347e: enq(MAGIC_DEBUG_KEY); return; /* F12 as debug hook */
        default: return;
        }
    default:
        return;
    }
}

static void enqueue_next_key(void) {
    uint64_t keyint = 0;
    ssize_t inlen = 0;
    byte c[7] = {0};
    int i;
    if (!key_available()) {
        return;
    }
    inlen = read(0, c, 7);
    if (inlen < 0) {
        printf("kb read failed\n");
        exit(1);
    }
    /* convert to integer */
    for (i = 0; i < inlen; i++) {
        keyint <<= 8;
        keyint |= c[i];
    }
    enqueue_single_key_with_mod(keyint, inlen);
}

int term_next_key(void) {
    int k;
    k = deq();
    if (k >= 0) {
        return k;
    }
    enqueue_next_key();
    k = deq();
    if (k >= 0) {
        return k;
    }
    return -1;
}

void term_set_raw(void) {
    struct termios new_settings;
    if (tcgetattr(0, &new_settings) < 0) {
        perror("tcgetattr failed");
        exit(1);
    }
    new_settings.c_lflag &= ~ICANON;
    new_settings.c_lflag &= ~ECHO;
    new_settings.c_lflag &= ~ISIG;
    new_settings.c_iflag &= ~IXON;
    new_settings.c_iflag &= ~IXOFF;
    new_settings.c_iflag &= ~ICRNL;
    if (tcsetattr(TCSADRAIN, 0, &new_settings) < 0) {
        perror("tcsetattr failed");
        exit(1);
    }
}

void term_init(void) {
    if (inited) {
        return;
    }
    if (tcgetattr(0, &orig_settings) < 0) {
        perror("tcgetattr failed");
        exit(1);
    }
    if (!isatty(1)) {
        fprintf(stderr, "stdout not a tty\n");
        exit(1);
    }
    inited = 1;
}

void term_reset(void) {
    size_t i;
    for (i = 0; mod_keys[i].len; i++) {
        mod_keys[i].state = MOD_NONE;
    }
    if (tcsetattr(TCSADRAIN, 0, &orig_settings) < 0) {
        perror("tcsetattr failed");
        exit(1);
    }
    fflush(stdout);
}

void term_set_cursor(byte row, byte col) {
    printf("\033[%hhu;%hhuH", row, col + 1);
}
