/*
This file is part of LazyPC.
Written by Taylor Gurney, 2022. Public domain.
*/

/* "fake" 8253 that uses cycle count instead of real clock time. */

#include <stdio.h>
#include <stdlib.h>

#include "clock.h"
#include "timer.h"

/* Intel 8253 Programmable Interval Timer (PIT) chip */

/* timer 0 tick */
static byte tick0;

/* BCD mode not implemented. */
static struct counter {
    word c;
    word maxval;
    byte latch_high_present;
    byte c_latch_high;
    byte latch_low_present;
    byte c_latch_low;
    byte do_flipflop;
    byte lsb_only;
    byte mode;
    uint64_t last_cycle_count;
} counter[3] = {0};

/*
enum counter_mode {
    terminal_interrupt = 0,
    oneshot = 1,
    rate = 2,
    square_wave = 3,
    sw_strobe = 4,
    hw_strobe = 5,
    rate_2 = 6,
    square_wave_2 = 7
};
*/

void timer_init(void) {
    int i;
    tick0 = 0;
    for (i = 0; i < 3; i++) {
        counter[i].c = 0;
        counter[i].maxval = 0;
        counter[i].latch_high_present = 0;
        counter[i].c_latch_high = 0;
        counter[i].latch_low_present = 0;
        counter[i].c_latch_low = 0;
        counter[i].do_flipflop = 0;
        counter[i].lsb_only = 1;
        counter[i].mode = 0;
        counter[i].last_cycle_count = 0;
    }
}

static void flip(struct counter* c) {
    if (c->do_flipflop) {
        c->lsb_only = (~c->lsb_only) & 1;
    }
}

static void refresh_timer(byte which) {
    struct counter* c;
    uint64_t intervals;
    uint64_t cycles_since;
    
    which &= 0x3;
    c = &counter[which];

    cycles_since = clock_get_cycle_count() - c->last_cycle_count;
    intervals = cycles_since / 4;
    if (!intervals) {
        /* nothing to do, not enough time elapsed */
        return;
    }
    c->last_cycle_count += intervals * 4;
    if (c->maxval) {
        intervals %= c->maxval;
    }
    intervals &= 0xffff;
    while (intervals) {
        intervals--;
        if (c->c == 0) {
            c->c = c->maxval;
        }
        c->c -= 1;
        if (!c->c) {
            if (!which) {
                /* timer tick. timer 0 connected to irq 0 */
                tick0 = 1;
            }
        }
    }
}

byte timer_read(byte which) {
    byte result;
    struct counter* c;
    which &= 0x3;
    refresh_timer(which);
    c = &(counter[which]);
    if (c->lsb_only) {
        if (c->latch_low_present) {
            result = c->c_latch_low;
            c->latch_low_present = 0;
        } else {
            result = (byte)c->c;
        }
    } else { /* MSB */
        if (c->latch_high_present) {
            result = c->c_latch_high;
            c->latch_high_present = 0;
        } else {
            result = (byte)((c->c & 0xff00) >> 8);
        }
    }
    flip(c);
    return result;
}

void timer_set_maxval(byte which, byte maxval) {
    struct counter* c;
    which &= 0x3;
    c = &(counter[which]);
    if (c->lsb_only) {
        c->maxval &= 0xff00;
        c->maxval |= maxval;
    } else {
        c->maxval &= 0x00ff;
        c->maxval |= (maxval << 8);
    }
    flip(c);
    c->c = c->maxval - 1;
}

void timer_set_mode(byte mode) {
    /* bit 0: 0=16 binary counter, 1=4 decade bcd counter
       bits 1-3: counter mode bits
       bits 4-5: read/write/latch format bits
       bits 6-7: counter select bits
    */
    struct counter* c;
    byte which = ((mode & 0xc0) >> 6) & 0x3;
    byte latchfmt = ((mode & 0x30) >> 4) & 0x3;
    byte modebits = ((mode & 0x0e) >> 1) & 0x7;
    
    if (which == 3) {
        fprintf(stderr, "8253 timer: got invalid counter selection\n");
        abort();
    }
    c = &(counter[which]);
    if (latchfmt == 0) {
        /* latch command. */
        c->c_latch_high = (c->c & 0xff00) >> 8;
        c->c_latch_low = (c->c & 0x00ff);
        c->latch_high_present = 1;
        c->latch_low_present = 1;
        return;
    }

    if (latchfmt == 1) {
        c->lsb_only = 1;
        c->do_flipflop = 0;
    } else if (latchfmt == 2) {
        c->lsb_only = 0;
        c->do_flipflop = 0;
    } else if (latchfmt == 3) {
        c->lsb_only = 1;
        c->do_flipflop = 1;
    }
    c->mode = modebits;
}

void timer_refresh(void) {
    refresh_timer(0);
    refresh_timer(1);
    refresh_timer(2);
}

void timer_dump_state(void) {
    int i;
    puts("+------------------------------------------------------+---------------+");
    timer_refresh();
    for (i = 0; i < 3; i++) {
        printf("| timer %i: %04hx (/ %04hx) [<%c%c> %02hhx%02hhx] [%c] [%c] [mode: %i]",
           i,
           counter[i].c,
           counter[i].maxval,
           (int)(counter[i].latch_high_present ? 'L' : ' '),
           (int)(counter[i].latch_low_present ? 'L' : ' '),
           counter[i].c_latch_high,
           counter[i].c_latch_low,
           (int)(counter[i].lsb_only ? '>' : '<'),
           (int)(counter[i].do_flipflop ? 'F' : ' '),
           counter[i].mode
           );
       if (i == 1) {
           printf(" | irq 0: [%c]    |\n", tick0 ? 'x' : ' ');
       } else {
           printf(" |               |\n");
       }
    }
}

byte timer_poll_timer_0(void) {
    timer_refresh();
    if (tick0) {
        tick0 = 0;
        return 1;
    }
    return 0;
}
